
#define _CRT_RAND_S
#include "pch.h"
#include "RNAStartOptimiser.h"

#ifdef _DEBUG
#pragma warning(disable:4244)
#endif


/*
	Optimises a set of PSSMs to acheive the best match between the calculated reads (model_reads) and
	the actual reads.  This needs a set of data that models the underlying (and unknown) RNA fragmentation 
	data.

	Options are to use regions from teh start where it is assumed that all the fragments starts are aligned
	or to use dat which has previously been computed using a previous guess as to the PSSM
*/


//#define COMPARE_ABS	//Compare abolsute difference between calculated and actual reads
//#define STORE_ABS_READS	//Store absolute read data

bool RNAStartDataSet::open(const string & filename,int numberOfNucs)
{

	if (!dataFile::open(filename))
		return false;
	int index = 0;
	while (readLine() && (index < 300))
	{
		resize(index+1);
		setData(at(index),numberOfNucs);
		index++;
	}
	return true;
};

bool RNAStartDataSet::open(const string & filename,mRNAinfoSet & genes,int Ngenes,bool reverse)
{
	if (!dataFile::open(filename))
		return false;
	readLine();
	int index = 0;
	string currentGene;
	mRNAinfo * currentGeneInfo = 0;

	while (readLine())
	{
		string gene = lineData(',');
		if (gene != currentGene)
		{
			index++;
			if (index > Ngenes)
			{
				break;
			}
			if ((currentGeneInfo = genes.findGeneInfo(gene)) == 0)
			{
				cout << "Failed to find gene " << gene << "\n";
				close();
				return false;
			}
			push_back(mRNAinfo(*currentGeneInfo));
			currentGene = gene;
			currentGeneInfo = &back();
			currentGeneInfo ->m_modelInfo.resize(currentGeneInfo->m_readData[0].reads.size());
			currentGeneInfo ->m_offset = 0;
		}
		int N = atoi(lineData(','));
		string c = lineData(',');
		if (c[0] != currentGeneInfo->m_sequence[N])
		{
			cout << "Chromosome mismatch\n";
			exit (1);
		}
		lineData(',');	//model
		float reads = (float)atof(lineData(','));
		if (reads)
			currentGeneInfo ->m_readData[reverse?1:0].reads[N] = log(LOG_OFFSET + reads)/Log2;	//reads
		else
			currentGeneInfo ->m_readData[reverse?1:0].reads[N] = DEFAULT_IF_NONE;
		currentGeneInfo ->m_modelInfo[N].m_breaks = (float)atof(lineData(','));	//breaks
	}

	close();
	return true;
};


void RNAStartDataSet::setData(mRNAinfo & data,int length)
{
	data.m_name = lineData(',');
	data.m_offset = atoi(lineData(','));
	data.m_readData[0].total = atoi(lineData(','));
	data.m_sequence = lineData(',');
	data.m_model.resize(length);
	data.m_modelInfo.resize(length);
	data.m_readData[0].reads.resize(length);
	data.m_readData[0].total = 0;
	for (int i = 0;i < length; i++)
	{
		int reads = atoi(lineData(','));
		data.m_readData[0].total += reads;
#ifdef STORE_ABS_READS
		data.m_readData[0].reads[i] = reads;
#else
		if (reads)
			data.m_readData[0].reads[i] = (float)log((double)reads + LOG_OFFSET)/Log2;
		else
			data.m_readData[0].reads[i] = DEFAULT_IF_NONE;
#endif
	}

	//	Assume that all the reads are as a result of breaks at the
	//	first location
	data.m_modelInfo[0].m_breaks = (float)data.m_readData[0].total;


}



RNAStartOptimiser::RNAStartOptimiser(const char * rootDir,
						 int window_size,int window_scale,int window_offset,int pssmOffset) :
		breakOptimiser(rootDir,0,0,-pssmOffset),
		m_window_size(window_size),m_window_scale(window_scale),m_window_offset(window_offset),
		m_first_gene(0), m_N_genes(0),
		m_carryOver(false)
{

}

bool RNAStartOptimiser::open(const string & filename,int numberOfNucs,bool logNorm)
{
	return (RNAdata.open(filename,numberOfNucs));
}

bool RNAStartOptimiser::open(const string & filename,mRNAinfoSet & genes,int Ngenes,bool logNorm)
{
	bool reverse = (strstr(filename.c_str(),"_ana_R_") != 0);

	if (RNAdata.open(filename,genes,Ngenes,reverse))
	{
		if (reverse)
			complementGenes();
		return true;
	}
	else
		return false;
}

void RNAStartOptimiser::complementGenes()
{
	for (size_t i = 0;i < RNAdata.size();i++)
	{
		RNAdata[i] = RNAdata[i].complement();
	}
}
void RNAStartOptimiser::getTopGenes(mRNAinfoSet & genes,int Ngenes,int firstGene,const string & firstGeneName)
{
	if (firstGeneName.empty())
		m_first_gene = firstGene;
	else
	{
		for (size_t i = 0;i < genes.topN.size();i++)
		{
			if (genes[genes.topN[i]].m_name == firstGeneName)
			{
				m_first_gene = (int)i;
				break;
			}
		}
	}
	m_N_genes = Ngenes;

	for (int i = m_first_gene;i < m_N_genes + m_first_gene;i++)
	{
		RNAdata.push_back(genes[genes.topN[i]]);
	}
}

void RNAStartOptimiser::generateSmoothedBreakInfo()
{
	for (size_t i = 0;i < RNAdata.size();i++)
	{
		RNAdata[i].generateSmoothedBreakInfo(m_window_size,m_window_scale,m_window_offset);
	}
}

void RNAStartOptimiser::outputScores(bool dataIsLogarithmic)
{
	char strCommand[MED_LINE_LENGTH];
	sprintf(strCommand, "%s \"%sresults*.csv\"\n", DEL_COM,m_rootDir.c_str());
	system(strCommand);

	char filename[200];
	sprintf(filename,"%sresults.csv",m_rootDir.c_str());
	FILE * f = fopen(filename,"w");
	if (f == 0)
	{
		cout << "Failed to open results file\n";
		return;
	}

	fprintf(f,"gene,total reads,start pos,pos,nuc,reads,model,error,PSSM A,PSSM B,breaks,outstanding,filteredBreaks\n");
	for (size_t i = 0;i < RNAdata.size(); i++)
	{
		double reads;
		for (size_t j = 0;j < RNAdata[i].m_model.size(); j++)
		{
#ifdef COMPARE_ABS
			if (RNAdata[i].m_model[j])
				reads = log(RNAdata[i].m_readData[0].reads[j] + LOG_OFFSET)/Log2;
			else
				reads = DEFAULT_IF_NONE;

#else
			reads = RNAdata[i].m_readData[0].reads[j];
#endif
			fprintf(f,"%s,%i,%i,%i,%c,%f,%f,%f,%i,%i,%f,%f,%f,%f\n",RNAdata[i].m_name.c_str(),
				RNAdata[i].m_readData[0].total,
				RNAdata[i].m_offset,j,RNAdata[i].m_sequence[j],
				reads,
				RNAdata[i].m_model[j][0],
				RNAdata[i].m_modelInfo[j].m_error,
				RNAdata[i].m_modelInfo[j].m_PSSM[0],RNAdata[i].m_modelInfo[j].m_PSSM[1],
				RNAdata[i].m_modelInfo[j].m_breaks,
				RNAdata[i].m_modelInfo[j].m_outstanding,
				RNAdata[i].m_model[j][1],
				RNAdata[i].m_model[j][2]);
		}
	}
	fprintf(f,"0\n");
	fclose(f);
}

#define HEIGHT_THRESH 4
#define DIFF_THRESH 1

void RNAStartOptimiser::outputPeaks()
{
	char filename[200];
	sprintf(filename,"%speaks.csv",m_rootDir.c_str());
	FILE * f = fopen(filename,"w");
	if (f == 0)
	{
		cout << "Failed to open file\n";
		cin >> filename;
	}

	fprintf(f,"index,gene,pos,reads,model,diff,sequence\n");
	int index = 1;
	for (size_t i = 0;i < RNAdata.size(); i++)
	{
		double reads,model;
		for (size_t j = 0;j < RNAdata[i].m_model.size(); j++)
		{
			reads = RNAdata[i].m_readData[0].reads[j];
			model = RNAdata[i].m_model[j][0];
			if ((reads > HEIGHT_THRESH) && ((reads - model) > DIFF_THRESH))
			{
				fprintf(f,"%i,%s,%i,%f,%f,%f,",index++,RNAdata[i].m_name.c_str(),j,reads,model,reads-model);
				for (int k = 0;(k < 20) && ((j + k) < RNAdata[i].m_model.size()); k++)
				{
					fprintf(f,"%c",tolower(RNAdata[i].m_sequence[j+k]));
				}
				fprintf(f,"\n");
				j += 4;
			}
		}

	}
	fclose(f);
}


ErrorPair RNAStartOptimiser::calculatePssmWeights()
{
	m_pssmBlocks.setGains();

	ErrorPair error;
	int PSSM_no = -1;

	//	RNAdata is an array with an entry for each of the genes that we are 
	//	optimising for.  Calculate the response for each in turn
	for (size_t i = 0; i < RNAdata.size(); i++)
	{
		RNAdata[i].initialiseModelArrays();
		error.NoWeightings += RNAdata[i].calculatePSSMWeights(m_pssmBlocks,false,0,m_carryOver,m_pssmOffset);
	}
	error.WithWeightings = error.NoWeightings;
	return error;
}


ErrorPair RNAStartOptimiser::ErrorFunc()
{
	m_pssmBlocks.ConvertXyzToPssm();
	ErrorPair Score = calculatePssmWeights();

	for (size_t block = 0; block < m_pssmBlocks.size(); block++)
	{

		for (size_t pssm = 0; pssm < m_pssmBlocks[block].size(); pssm++)
		{
			//	This bits stops the optimiser coming up with pssms with negative cooefficients
			if((pssm == 0) || (m_pssmBlocks[block][pssm].m_scalers.size() == 0))
			{
				pssmVector pv;
				for (size_t i = 0; i <  m_pssmBlocks[block][pssm].size(); i++)
				{
					pv.ConvFrom(m_pssmBlocks[block][pssm].m_xyzWeights[i]);

#ifdef _DEBUG
					pssmDataType & pd = m_pssmBlocks[block][pssm];
#endif
					for (int j = 0; j < 4; j++)
					{
#ifdef LINEAR_PSSM
						if (pv[j].value < 0)
							Score.WithWeightings -= (pv[j].value * 100);
						if (pv[j].value > 1)
							Score.WithWeightings += ((pv[j].value -1) * 100);
#else
						if (pv[j].value < -100)
							Score.WithWeightings -= 10000 * (pv[j].value + 100);
#endif
					}
				}
			}
			//	And this bit stops it producing scalers that are outside the range 0-1.
			for (size_t i = 0; i < m_pssmBlocks[block][pssm].m_scalers.size(); i++)
			{
				if(m_pssmBlocks[block][pssm].m_scalers[i].value > 1)
					Score.WithWeightings += 10*(m_pssmBlocks[block][pssm].m_scalers[i].value-1);
				else if(m_pssmBlocks[block][pssm].m_scalers[i].value < 0)
					Score.WithWeightings -= 10*(m_pssmBlocks[block][pssm].m_scalers[i].value);
			}
		}
	}
	return Score;
}

void RNAStartOptimiser::SaveResults(bool toFile)
{
	cout << "\n" << m_totalIterations << "," << m_ErrLast.WithWeightings << "," << m_ErrLowest.WithWeightings << "\n";
	if (toFile)
	{
		outputWeights(m_ErrLast);
		SaveProgress();
	}
}

void RNAStartOptimiser::outputWeightParameters(FILE * f)
{
	fprintf(f,"[parameters]\n");
	fprintf(f,"window size,%i\n",m_window_size);
	fprintf(f,"window scaling,%i\n",m_window_scale);
	fprintf(f,"window offset,%i\n",m_window_offset);
	fprintf(f,"start gene,%i\n",m_first_gene);
	fprintf(f,"carry over,%c\n",m_carryOver?'y':'n');
	fprintf(f,"N genes,%i\n",m_N_genes);
	fprintf(f,"PSSM Offset,%i\n",m_pssmOffset);
}

void RNAStartOptimiser::readWeightParameters(std::ifstream & f,mystring & buffer)
{
	while (getDataLine(f,buffer))
	{
		const char * p;
		if (buffer.startsWith("window size,",p))
			m_window_size = atoi(p);
		else if (buffer.startsWith("window scaling,",p))
			m_window_scale = atoi(p);
		else if (buffer.startsWith("window offset,",p))
			m_window_offset = atoi(p);
		else if (buffer.startsWith("carry over,",p))
			m_carryOver = (*p == 'y');
		else if (buffer.startsWith("PSSM Offset,",p))
			m_pssmOffset = atoi(p);
	}
};


