#include "pch.h"
#include <string>
#include <vector>
#include <map>
#include <cstdio>
#include <iostream>
#include <algorithm>
#include "barFile.h"

barFile::barFile() :m_pBARData(0),m_file(0)
{
}

barFile::~barFile()
{
	close();
}

void barFile::close()
{
	if (m_pBARData)
		Affy_BARData_Destroy(&m_pBARData);
	m_pBARData = 0;

	if (m_file)
	{
		fclose(m_file);
		m_file = 0;
	}

	clear();
	m_data.clear();
}

bool barFile::open(const string & file_name,bool fast)
{
	close();
	m_filename = file_name;
	if (fast)
	{
		m_bar = 0;
		char fileName[MED_LINE_LENGTH];
		strcpy(fileName,file_name.c_str());
		m_pBARData = Affy_LoadBAR_Fast(fileName);

		for(int ni=0; ni<m_pBARData->nSeqNum; ni++)
		{
			if(m_pBARData->vSeqData[ni]->nDataNum <= 0)
				continue;
			seq_data sequence;
			sequence.data.num_data = m_pBARData->vSeqData[ni]->nDataNum;
			sequence.data.name = m_pBARData->vSeqData[ni] ->pSeqName->m_pString;
			sequence.file_offset = 0;
			sequence.tagBarData = m_pBARData->vSeqData[ni];
			m_sequences.insert (sequence_map::value_type(sequence.data.name,sequence));
		}

	}
	return true;

};




vector<string> barFile::GetChrs()
{
	vector<string> chrs;

	if (m_pBARData)
	{
		for(int ni=0; ni<m_pBARData->nSeqNum; ni++)
			chrs.push_back(m_pBARData->vSeqData[ni]->pSeqName->m_pString);
	}
	else
	{
		for (vector<bar_sequence>::const_iterator i = m_bar ->sequences.begin();i != m_bar ->sequences.end();i++)
			chrs.push_back((*i).name);
	}
	return chrs;
}

//	Sets the length of one of the sets of data in a bar file, deleting any previous data
bool barFile::setNEntries(const string & chr,int p)
{
	seq_data & seq = (*m_sequences.find(chr)).second;

	if (m_pBARData)
	{
		seq.tagBarData ->nDataNum = p;

		for(int nj=0; nj<seq.tagBarData -> nColNum; nj++)
		{
			DestroyDoubleMatrix(seq.tagBarData ->vData[nj]);
			if (seq.tagBarData ->nDataNum)
			{
				seq.tagBarData ->vData[nj] = CreateDoubleMatrix(1, seq.tagBarData ->nDataNum);
				if(seq.tagBarData -> vData[nj] == NULL)
				{
					printf("Error: Affy_LoadBAR, cannot allocate memory for loading sequence data.\n");
					exit(EXIT_FAILURE);
				}
			}
			else
				seq.tagBarData ->vData[nj] = 0;
		}
		return true;
	}
	else
	{
		return false;
	}
}

bool barFile::getNext(const string & chr,chrData & data, size_t & p)
{

	if (m_pBARData)
	{
		seq_data & seq = (*m_sequences.find(chr)).second;
		if (p >= (uint)seq.tagBarData ->nDataNum)
			return false;

		data.value = (float)seq.tagBarData ->vData[1] ->pMatElement[p];
		double loc = seq.tagBarData ->vData[0] ->pMatElement[p++];
		if (loc >= 0)
			data.location = (size_t)loc;
		else
			data.location = 0;
	}
	else
	{
		return false;
	}
	return true;
}
bool barFile::getAfter(const string & chr,size_t index,chrData & data, size_t & p)
{

	if (m_pBARData)
	{
		seq_data & seq = (*m_sequences.find(chr)).second;
		size_t min = 0;
		size_t max = seq.tagBarData ->nDataNum;
		do
		{
			p = (min + max)/2;
			data.location = (size_t)seq.tagBarData ->vData[0] ->pMatElement[p];
			if ((data.location == index) || ((max - min) == 1))
			{
				if ((max - min) == 1)
				{
					if (p >= ((uint)seq.tagBarData ->nDataNum -1))
						return false;
					data.location = (int)seq.tagBarData ->vData[0] ->pMatElement[++p];
				}

				data.value = (float)seq.tagBarData ->vData[1] ->pMatElement[p++];
				return true;
			}
			else if (data.location > index)
				max = p;
			else
				min = p;
		} while (true);
	};
	return false;
}



bool barFile::normalise(const string & chr,unsigned int p,double normaliser)
{

	if (m_pBARData)
	{
		seq_data & seq = (*m_sequences.find(chr)).second;
		if (p >= (uint)seq.tagBarData ->nDataNum)
			return false;

		seq.tagBarData ->vData[1] ->pMatElement[p] *= normaliser;
		return true;
	}
	return false;

}

seq_data & barFile::getSeqData(const string & chromosome)
{
	seq_data & seq = (*m_sequences.find(chromosome)).second;
	return seq;
}


int barFile::getNEntries(const string & chr)
{
	if (m_pBARData)
	{
		seq_data & seq = (*m_sequences.find(chr)).second;
		return seq.tagBarData ->nDataNum;
	}
	else
	{
		map<string,int>::const_iterator i = m_bar ->seq_map.find(chr);
		if (i == m_bar ->seq_map.end())
			return 0;

		return (m_bar -> sequences[(*i).second].num_data);
//		return seq.data.num_data;
	}
}




size_t barFile::getLength(const string & chr)
{
	size_t maxBar = getNEntries(chr)-1;
	chrData data;
	getNext(chr,data,maxBar);
	return data.location;
}

bool barFile::Initialise(vector<string> & seqNames)
{
	size_t nSeqNum = seqNames.size();
	int nColNum = 1;

	/* create BAR object */
	m_pBARData = Affy_BARData_Create();
	if(m_pBARData == NULL)

	{
		printf("Error: HTS_Aln2BAR, BARData object was not created!\n");
		exit(EXIT_FAILURE);
	}
	strcpy(m_pBARData->strMagicnumber, "barr\r\n\032\n");
	m_pBARData->fVersionnumber = 2.0;
    m_pBARData->nSeqNum = (int)nSeqNum;
	m_pBARData->nColNum = nColNum+1;
	m_pBARData->nParamNum = 0;
	m_pBARData->vParamName = NULL;
	m_pBARData->vParamValue = NULL;
	m_pBARData->pFieldType = CreateIntMatrix(1, m_pBARData->nColNum);
	if(m_pBARData->pFieldType == NULL)
	{
		printf("Error: HTS_Aln2BAR, cannot allocate memory for field type.\n");
		exit(EXIT_FAILURE);
	}
	m_pBARData->pFieldType->pMatElement[0] = 2;
	int ni;
	for(ni=1; ni<m_pBARData->nColNum; ni++)
		m_pBARData->pFieldType->pMatElement[ni] = 1;

	m_pBARData->vSeqData = (struct tagBARSeq **)calloc(m_pBARData->nSeqNum, sizeof(struct tagBARSeq *));
	if(m_pBARData->vSeqData == NULL)
	{
		printf("Error: TileMapv2_TXT2BAR, cannot allocate memory for loading sequence data.\n");
		exit(EXIT_FAILURE);
	}

	for(ni=0; ni<m_pBARData->nSeqNum; ni++)
	{
		/* create BARSeq object */
		m_pBARData->vSeqData[ni] = Affy_BARSeq_Create();
		if(m_pBARData->vSeqData[ni] == NULL)
		{
			printf("Error: TileMapv2_TXT2BAR, cannot create BARSeq object.\n");
			exit(EXIT_FAILURE);
		}

		m_pBARData->vSeqData[ni]->pSeqGroupName = NULL;
		m_pBARData->vSeqData[ni]->pSeqVersion = NULL;
		m_pBARData->vSeqData[ni]->nParamNum = 0;
		m_pBARData->vSeqData[ni]->vParamName = NULL;
		m_pBARData->vSeqData[ni]->vParamValue = NULL;

		m_pBARData->vSeqData[ni]->nColNum = m_pBARData->nColNum;

//		chp2 = strchr(chp1, '\t');
//		*chp2 = '\0';
//		strcpy(strTemp, chp1);
		char strTemp[100];
		strcpy(strTemp,seqNames[ni].c_str());

		StringAddTail(&(m_pBARData->vSeqData[ni]->pSeqName), strTemp);
//		chp1 = chp2+1;

//		chp2 = strchr(chp1, '\t');
//		*chp2 = '\0';
//		strcpy(strTemp, chp1);
//		m_pBARData->vSeqData[ni]->nDataNum = atoi(strTemp);
		m_pBARData->vSeqData[ni]->nDataNum = 0;
//		chp1 = chp2+1;


		m_pBARData->vSeqData[ni]->vData = (struct DOUBLEMATRIX **)calloc(m_pBARData->vSeqData[ni]->nColNum, sizeof(struct DOUBLEMATRIX *));
		if(m_pBARData->vSeqData[ni]->vData == NULL)
		{
			printf("Error: HTS_Aln2BAR, cannot allocate memory for loading bpmap coordinate data!\n");
			exit(EXIT_FAILURE);
		}
/*		if(m_pBARData->vSeqData[ni]->nDataNum > 0)
		{
			for(int nj=0; nj<m_pBARData->vSeqData[ni]->nColNum; nj++)
			{
				m_pBARData->vSeqData[ni]->vData[nj] = CreateDoubleMatrix(1, m_pBARData->vSeqData[ni]->nDataNum );
				if(m_pBARData->vSeqData[ni]->vData[nj] == NULL)
				{
					printf("Error: HTS_Aln2BAR, cannot allocate memory for loading bpmap coordinate data!\n");
					exit(EXIT_FAILURE);
				}
			}
		}
*/
		seq_data sd;
		sd.file_offset = 0;
		sd.tagBarData = m_pBARData->vSeqData[ni];


		m_sequences.insert(sequence_map::value_type(seqNames[ni],sd));
	}


	return true;
}


bool barFile::save(const string & filename)
{
	/* write to BAR file */

	char strBARFile[MED_LINE_LENGTH];
	strcpy(strBARFile,filename.c_str());

	struct INTMATRIX * pCol = CreateIntMatrix(1,m_pBARData->nColNum);
	if(pCol == NULL)
	{
		printf("Error: HTS_Aln2BAR, cannot create output column information!\n");
		exit(EXIT_FAILURE);
	}
	
	pCol->pMatElement[0] = 1;
	pCol->pMatElement[1] = 1;
	Affy_SaveBAR_Columns_Fast(strBARFile, m_pBARData, pCol);
	DestroyIntMatrix(pCol);
	/* return */
	return true;

}

