#include "pch.h"
#include "genomeFile.h"

#define FASTA_BUFFER_LENGTH 1000000

string sequence::complement()
{
	string s;
	int z = (int)size();
	for (int i = z-1; i >= 0; i--)
	{
		switch (at(i))
		{
		case ('A'): case ('a'):s.push_back('T'); break;
		case ('C'): case ('c'):s.push_back('G'); break;
		case ('G'): case ('g'):s.push_back('C'); break;
		case ('T'): case ('t'):s.push_back('A'); break;
		case ('N'): s.push_back('N'); break;
		}
	}
	return s;
}



genomeFile::genomeFile(void) :m_pSeq(0),m_seqData(0),m_vChrName(0),m_nChrNum(0)
{
}

genomeFile::~genomeFile(void)
{
	SequenceDelete(m_pSeq);
	freeChromsomeData();
}
void genomeFile::freeChromsomeData()
{
	if (m_nChrNum)
	{
		DestroyIntMatrix(m_pChrLen);
		for(int ni=0; ni<m_nChrNum; ni++)
		{
			DeleteString(m_vChrName[ni]);
			m_vChrName[ni] = NULL;
		}
		free(m_vChrName);
	}
	m_nChrNum = 0;
}


void genomeFile::setRoot(const string & root)
{
	char strLine[MED_LINE_LENGTH];

	strcpy(m_root,root.c_str());

	char fname[LINE_LENGTH];
	sprintf(fname,"%s%cchrlen.txt",m_root,FILE_SEP);
	m_pChrLen = IMLOAD(fname);
	if(m_pChrLen == NULL)
	{
		printf("Error: Genome_CreateHash_Main, cannot load chromosome length!\n");
		exit(EXIT_FAILURE);
	}
	m_nChrNum = m_pChrLen->nHeight;

	sprintf(fname, "%s%cchrlist.txt",m_root,FILE_SEP);
	FILE * f = fopen(fname, "r");
	if(f == NULL)
	{
		printf("Error: Genome_CreateHash_Main, cannot load chromosome name!\n");
		exit(EXIT_FAILURE);
	}

	int ni = 0;
	m_vChrName = (struct tagString **)calloc(m_nChrNum, sizeof(struct tagString *));

	while(fgets(strLine, MED_LINE_LENGTH, f)!=NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] == '\0')
			continue;

		if(ni >= m_nChrNum)
		{
			printf("Error: Genome_CreateHash_Main, chromosome number not match!\n");
			exit(EXIT_FAILURE);
		}

		StringAddTail(m_vChrName+ni, strLine);
		ni++;
	}

}

size_t genomeFile::NChrs()
{
	return m_nChrNum;
}

size_t genomeFile::getChrLength(size_t chrNum)
{
	if (m_seqData)
	{
		return m_length;
	}
	else
	{
		return (unsigned long)IMGETAT(m_pChrLen,(long)chrNum,0);
	}
}

unsigned long long genomeFile::getTotalLength()
{
	unsigned long long length = 0;
	for (size_t i = 0;i < NChrs(); i++)
	{
		length += getChrLength(i);
	}
	return length;
}


const char * genomeFile::getChrName(size_t chrNum)
{
	if ((chrNum & 0x3f) < NChrs())
		return m_vChrName[(chrNum & 0x3f)]->m_pString;
	else
		return "XXX";
}


const char * genomeFile::getseq(const char * chr, size_t start, size_t end)
{
	if (m_seqData)
	{
		m_seqBuffer.assign(m_seqData+start,end-start+1);
		return m_seqBuffer.c_str();
	}
	else
	{
		SequenceDelete(m_pSeq);

		char filename[LINE_LENGTH];
		sprintf(filename,"%s%c%s.sq",m_root,FILE_SEP,chr);

		if (end > start)
			m_pSeq = Genome_Code_4bit_GetSeq(filename, (int)start, (int)end);
		else
			m_pSeq = Genome_Code_4bit_GetCompSeq(filename,(int)start, (int)end);
		if(m_pSeq == NULL) {
			//		error_msg = "error reading sequence file.\n";
			throw(1);
		}
		return m_pSeq->m_pSequence->m_pString;
	}
}

bool genomeFile::open(const string & filename)
{
	FILE * f = fopen(filename.c_str(),"rt");
	if (f == 0)
		return false;

	char status[MED_LINE_LENGTH];
	m_length = 0;

	int bufferLength = FASTA_BUFFER_LENGTH;

	m_seqData = (char *)malloc(bufferLength);
	char * buffPointer = m_seqData;
	fgets(status,MED_LINE_LENGTH,f);
	if (status[0] == '>')
		m_genStatus = status;
	else
	{
		fseek(f,0,SEEK_SET);
	}

	size_t chars = strlen(fgets(buffPointer,MED_LINE_LENGTH,f))-1;
	
	while (!feof(f))
	{
		char * p = 0;

		//	Some lines may have numbers at the end.  Ignore these
		size_t nucs = strspn(buffPointer,"ATCG");
		m_length += nucs;
		buffPointer += nucs;

		//	Increment buffer in FASTA_BUFFER_LENGTH increments
		if (bufferLength - m_length < MED_LINE_LENGTH)
		{
			bufferLength += FASTA_BUFFER_LENGTH;
			m_seqData = (char *)realloc(m_seqData,bufferLength);
			buffPointer = m_seqData + m_length;
		}
		p = fgets(buffPointer,MED_LINE_LENGTH,f);
		if (p)
			chars = strlen(p);
	}
	size_t nucs = strspn(buffPointer,"ATCG");
	m_length += nucs;
	buffPointer[nucs]=0;

	//	Trim the buffer to what is needed
	m_seqData = (char *)realloc(m_seqData,m_length+1);

	return true;
}


void genomeFile::printFasta(FILE * f,const char * title,size_t start,size_t end)
{
	fprintf(f,"> %s\n",title);
	while(end > start)
	{
		size_t len = min<size_t>(end-start,80);
		//	getset is from start to end inclusive
		string seq = getseq("",start,start+len-1);
		fprintf(f,"%s\n",seq.c_str());
		start += len;
	}
}

