#include "pch.h"
#include "histData.h"

/*****************************************************

histData class

*****************************************************/


histData::histData(int size) : vector<double>(size)
		,m_nHists(size)
		,m_dataSets(0)
		,m_dataSum(0)
{
	for (int i = 0; i < m_nHists; i++) at(i) = 0;
};
void histData::inc(int bin,int amount)
{

	if ((bin >= 0) && (bin < m_nHists))
	{
		at(bin)+= amount;
		m_dataSum += amount;
	}
};

double histData::addData(const barFile & barData,int reference)
{
	//	bardata contains the list of the breaks (magnitude and location) within this region
	double sum = 0;
	for (size_t k = 0; k < barData.size(); k++) {
		const float & v = barData[k][1].data_float;
		at(barData[k][0].data_int - reference) += v;
		m_dataSum += v;
		}

	m_dataSum += sum;
	m_dataSets++;

	return sum;
}



void histData::average()
{
	if (m_dataSets == 0) return;

	for (size_t i = 0; i < size(); i++)
		at(i) /= m_dataSets;
};

void histData::normalise()
{
	if (m_dataSum == 0) return;

	double average = m_dataSum/size();

	for (size_t i = 0; i < size(); i++)
		at(i) /= average;
};


/*****************************************************

histDataVec class

*****************************************************/

void histDataVec::average()
{
	for (size_t i = 0; i < size(); i++)
		at(i).average();
};

void histDataVec::normalise()
{
	for (size_t i = 0; i < size(); i++)
		at(i).normalise();
};

void histDataVec::combine(int vec1,int vec2,bool forwardStrand,bool reverseStrand)
{
	//	Combince together the data associated with the two directions.

	//	Forward and Reverse strand parameters allow just one direction or the other to be analysed
	size_t size = at(vec1).m_nHists;
	at(vec1).m_dataSum = 0;
	double sum = 0;
	int divisor = 0;
	double dataSets = 0;

	if (forwardStrand)
	{
		divisor++;
		dataSets += at(vec1).m_dataSets;
	}
	if (reverseStrand)
	{
		divisor++;
		dataSets += at(vec2).m_dataSets;
	}

	if (dataSets != 0)
	{
		for (size_t i = 0;i < size; i++)//original
		{
			sum = 0;

			if (forwardStrand) sum += at(vec1)[i];
			if (reverseStrand) sum += at(vec2)[size - i - 1];

			at(vec1)[i] = sum/dataSets;
			at(vec1).m_dataSum += at(vec1)[i];
		}

		double normaliser = at(vec1).m_dataSum / size;

		if (normaliser != 0)	//There may be no breaks associated with the binding region
		{
			for (size_t i = 0;i < size; i++)//original
			{
				at(vec1)[i] /= normaliser;
			}
		}
	}

}
/*****************************************************

histDataArray class

*****************************************************/

void histDataArray::normalise()
{
	for (size_t i = 0; i < size(); i++)
		at(i)[0].normalise();
};

//	Adds some barFile type data giving the number of sequence reads through a region
//	to the hostograms
double histDataArray::addData(const barFile & barData,int vecOffset,int pssm_score_group,int reference)
{
	return (*this)[pssm_score_group][vecOffset].addData(barData,reference);
}



size_t histDataArray::combine(int vec1,int vec2,
							  bool forwardStrand,bool reverseStrand,
							  int combineThreshold)
{
	for (int i = (int)size() - 1;i >= 0; i--)
	{
		int dataSets = at(i)[vec1].m_dataSets + at(i)[vec2].m_dataSets;
		if (dataSets != 0)
		{
			if ((dataSets < combineThreshold) && (i > 0))
			{
				for (size_t j = 0;j < at(i)[vec1].size();j++)
				{
					at(i-1)[vec1][j] += at(i)[vec1][j];
					at(i-1)[vec2][j] += at(i)[vec2][j];
					at(i)[vec1][j] = 0;
					at(i)[vec2][j] = 0;
				}
				at(i-1)[vec1].m_dataSets += at(i)[vec1].m_dataSets;
				at(i-1)[vec2].m_dataSets += at(i)[vec2].m_dataSets;
				at(i-1)[vec1].m_dataSum += at(i)[vec1].m_dataSum;
				at(i-1)[vec2].m_dataSum += at(i)[vec2].m_dataSum;
				at(i)[vec1].m_dataSets = 0;
				at(i)[vec2].m_dataSets = 0;
				at(i)[vec1].m_dataSum = 0;
				at(i)[vec2].m_dataSum = 0;
			}
			else
			{
				at(i).combine(vec1,vec2,forwardStrand,reverseStrand);
			}
		}
	}

	return 0;
}



void histDataArray::output(const string & filename,int vec,int width,const string & cons,
						   const char ** titles)
{
	FILE * output = fopen(filename.c_str(),"w");
	size_t depth = size();

	if (output == NULL)
	{
		std::cout << "Unable to open output file";
	}
	else
	{
		fprintf(output,",");
		size_t j = 0;
		while (titles[j])
			fprintf(output,", %s",titles[j++]);
		fprintf(output,"\n");

		for (j = 0; j < depth; j++)
			fprintf(output,",%e",at(j)[vec].m_dataSum);
		fprintf(output,"\n");
		for (int i = 0; i < width;i++)
		{
			if ((int)cons.size() > i)
				fprintf(output,"%c",cons[i]);
			for (j = 0; j < depth; j++)
				fprintf(output,",%e",at(j)[vec][i]);
			fprintf(output,"\n");
		}
	}
}

string seqData::consensus()
{
	string cons;
	vector<vector<int> > vals(at(0).size(),vector<int>(4));

	//	Count up the number of each nucleotide in each position
	for (size_t i = 0;i < size();i++)
	{
		for (size_t j = 0;j < at(i).size();j++)
		{
			switch (at(i)[j])
			{
			case 'A':
			case 'a':
				vals[j][0]++;
				break;
			case 'C':
			case 'c':
				vals[j][1]++;
				break;
			case 'G':
			case 'g':
				vals[j][2]++;
				break;
			case 'T':
			case 't':
				vals[j][3]++;
				break;
			}
		}
	}

	//	And then find the most frequenctly occurring, and see if it exceeds the threshold of
	//	0.5
	for (size_t i = 0; i < at(0).size();i++)
	{
		int max = 0;
		char c;
		for (int j = 0;j < 4;j++)
		{
			if (vals[i][j] > max)
			{
				max = vals[i][j];
				switch (j)
				{
				case 0: c = 'A';break;
				case 1: c = 'C';break;
				case 2: c = 'G';break;
				case 3: c = 'T';break;
				}
			}
		}
		if (max > 0.5 * size())
			cons += c;
		else
			cons += '-';
	}

	return cons;
}

#define MAX_SEQUENCES 300000

void seqData::addSequence(const char * seq,int & count)
{
	m_Nseqs++;
	if (count < MAX_SEQUENCES)
	{
		push_back(seq);
		count++;
	}
};


void seqData::CombineWithReverseStrands(const seqData & revStrands,size_t offset)
{
	int count = 0;
	if (revStrands.size())
	{
		size_t full_strSize = revStrands[0].size();
		size_t final_strSize = full_strSize;

		char * seq = (char *)malloc(sizeof(char) * (full_strSize+1));

		for (size_t i = 0;i < revStrands.size();i++)
		{
			size_t j;

//			for (j = 0; j < maxOffset; j++)
//				seq[j] = 'X';

			for (j = 0; j < full_strSize-2; j++)
			{
				switch (revStrands[i][full_strSize-j-offset-1])
				{
				case 'A':case 'a': seq[j] = 'T'; break;
				case 'C':case 'c': seq[j] = 'G'; break;
				case 'G':case 'g': seq[j] = 'C'; break;
				case 'T':case 't': seq[j] = 'A'; break;
				}
			}
			seq[j] = 0;
			addSequence(seq,count);
		}
	}
}
