
#include "pch.h"
#include "breakData.h"
#include "breakOptimiser.h"
#include "pssmData.h"

#ifdef _DEBUG
#pragma warning (disable: 4244 4305)
#endif


orderedList::orderedList(size_t size):m_length(0),m_allocated(size) 
	{
		m_data = (DOUBLE_TYPE * )malloc(m_allocated * sizeof(DOUBLE_TYPE));
#ifdef ORDERED_LIST_DEBUG
		m_debugData.resize(m_allocated);
#endif
	};

//	An explicit copy constructor because the default one does not copy the data pointed to 
//	by m_data
orderedList::orderedList(const orderedList & ol)
{
	m_length = ol.m_length;
	m_allocated = ol.m_allocated;
	m_data = (DOUBLE_TYPE * )malloc(sizeof(ol.m_data));
	memcpy(m_data,ol.m_data,sizeof(ol.m_data));

}

void orderedList::insert(DOUBLE_TYPE v)
	{
		if(m_length > m_allocated)
			throw;
		for (size_t i = 0;i < m_length;i++)
		{
			if (v < m_data[i])
			{
				memmove(&m_data[i+1],&m_data[i],(m_length-i) * sizeof(DOUBLE_TYPE));
				m_data[i] = v;
				m_length++;
#ifdef ORDERED_LIST_DEBUG
				for (size_t j = 0;j < m_length; j++)
					m_debugData[j] = m_data[j];
#endif
				return;
			}
		}
		m_data[m_length++] = v;
#ifdef ORDERED_LIST_DEBUG
		for (size_t j = 0;j < m_length; j++)
			m_debugData[j] = m_data[j];
#endif
		return;
	}


void orderedList::resize(size_t size)
{
	m_length = 0;
	if (size != m_allocated)
	{
		m_data = (DOUBLE_TYPE * )realloc(m_data,size * sizeof(DOUBLE_TYPE));
#ifdef ORDERED_LIST_DEBUG
		m_debugData.resize(size);
#endif
		m_allocated = size;
	};
}


/**************************************************************************************
pssmVector the class that holds the pssm data in xyz format during optimisation
if LINEAR_PSSM then the values are just held linearly
otherwise
The values are held internally as the log of the value. Negative numbers are held internally as
teh number shifted by -100,
**************************************************************************************/

DOUBLE_TYPE pssmVector::maxValue() const
{
	DOUBLE_TYPE m = -100;
	for (size_t i = 0; i < 4;i++)
		if (at(i).value > m) m =at(i).value;
	return m;
}

DOUBLE_TYPE pssmVector::infoContent() const
{
	DOUBLE_TYPE res = 0;
	for (size_t i = 0; i < 4;i++)
		 res += (log(4 * posValue(i))/log(2.0) * posValue(i));
	return res;
}

DOUBLE_TYPE pssmVector::absValue(size_t i) const
{
#ifdef LINEAR_PSSM
	return abs(at(i).value);
#else
	if (at(i).value <= -100)
		return (at(i).value + 100);
	return exp(at(i).value);
#endif
}
DOUBLE_TYPE pssmVector::posValue(size_t i) const
{
#ifdef LINEAR_PSSM
	return at(i).value;
#else
	if (at(i).value < -100)
		return 0.000001;
	return exp(at(i).value);
#endif
}

void pssmVector::setValue(size_t i,DOUBLE_TYPE value)
{
#ifdef LINEAR_PSSM
	at(i).value = value;
#else
	if (value <= 0)
		at(i).value = -100 + value;
	else
		at(i).value = log(value);
#endif
}

void pssmVector::makeComplement()
{
	DOUBLE_TYPE a = at(0).value;
	at(0).value = at(3).value;
	at(3).value = a;

	a = at(1).value;
	at(1).value = at(2).value;
	at(2).value = a;
}

void pssmVector::print(FILE * f)
{
	for (size_t i = 0; i < size();i++)
		fprintf(f,"%0.4f ",posValue(i));
	fprintf(f,"\n");
}
void pssmVectorVector::print(FILE * f,int offset)
{
	fprintf(f,"#FontSize %i\n",PSSM_FONT_SIZE);
	fprintf(f,"#labels");
	bool first = true;
	for (int i = 0; i < (int)size();i++)
	{
		if (first)
			first = false;
		else
			fprintf(f,",");

		if (offset + i >= 0) 
			fprintf(f,"%i",i+offset + 1);
		else
			fprintf(f,"%i",i+offset);
	}
	fprintf(f,"\n");

	for (size_t i = 0; i < size();i++)
		at(i).print(f);
}


void pssmVector::ConvFrom(const xyzVector & xv,DOUBLE_TYPE scaler)
{
	//	Creates a pssmVector, which holds the values from an xyzVector
	//	These may be log or linear represenation of the data, depending on the mode
#ifdef MAX_EQU_1
	DOUBLE_TYPE N = 1.0/max<DOUBLE_TYPE>((1-xyzW[0].value)*(fabs(xyzW[1].value) + 1),(1+xyzW[0].value)*(fabs(xyzW[2].value) + 1));
#else
	DOUBLE_TYPE N = 0.25;
#endif
//	if (scaler > 1) 
//		scaler = 1;;

#ifdef SIMPLE_MAPPING
	setValue(0,(xv[0].value-N) * scaler +N);
	setValue(1,(xv[1].value-N) * scaler +N);
	setValue(2,(xv[2].value-N) * scaler +N);
	setValue(3,1.0 - absValue(0) -absValue(1) - absValue(2));
#else
	setValue(0,N*(1+xv[1].value * scaler)*(1-xv[0].value * scaler));
	setValue(1,N*(1+xv[2].value * scaler)*(1+xv[0].value * scaler));
	setValue(2,N*(1-xv[2].value * scaler)*(1+xv[0].value * scaler));
	setValue(3,N*(1-xv[1].value * scaler)*(1-xv[0].value * scaler));
#endif
	setOptimising(xv[0].optimise);
}

void pssmVectorVector::ConvFrom(const xyzVectorVector & pv,DOUBLE_TYPE scale)
{
	resize(pv.size());
	for (size_t i = 0;i < size(); i++)
		at(i).ConvFrom(pv[i],scale);
}

pssmVector & pssmVector::operator *= (DOUBLE_TYPE a)
{
	for (size_t i = 0; i < size(); i++)
#ifdef LINEAR_PSSM
		at(i).value *= a;
#else
		setValue(i,absValue(i)*a);
#endif
	return *this;
}

pssmVectorVector & pssmVectorVector::operator *= (DOUBLE_TYPE a)
{
	for (size_t i = 0; i < size(); i++)
		at(i) *= a;
	return *this;
}

pssmVector & pssmVector::operator += (const pssmVector & a)
{
	for (size_t i = 0; i < size(); i++)
#ifdef LINEAR_PSSM
		at(i).value += a[i].value;
#else
		setValue(i,absValue(i)+a.absValue(i));
#endif
	return *this;
}

pssmVectorVector & pssmVectorVector::operator += (const pssmVectorVector & a)
{
	for (size_t i = 0; i < size(); i++)
		at(i) += a[i];
	return *this;
}



/**************************************************************************************
xyzVector the class that holds the pssm data in xyz format during optimisation
**************************************************************************************/

void xyzVector::ConvFrom(const pssmVector &pv,DOUBLE_TYPE scale)
{
#ifdef SIMPLE_MAPPING
	double N = 0.25;
	at(0).value = (pv.absValue(0) - N) * scale + N;
	at(1).value = (pv.absValue(1) - N) * scale + N;
	at(2).value = (pv.absValue(2) - N) * scale + N;

#else
	DOUBLE_TYPE a = pv.absValue(0);
	DOUBLE_TYPE c = pv.absValue(1);
	DOUBLE_TYPE g = pv.absValue(2);
	DOUBLE_TYPE t = pv.absValue(3);

	at(0).value = (1 - 2 * (a+t)/(a+c+g+t)) * scale;
	at(1).value = (1 - 2 * t/(a + t)) * scale;
	at(2).value = (1 - 2 * g/(c + g)) * scale;
#endif
	setOptimising(pv[0].optimise);
};

void xyzVectorVector::ConvFrom(const pssmVectorVector & pv,DOUBLE_TYPE scale)
{
	resize(pv.size());
	for (size_t i = 0;i < size(); i++)
		at(i).ConvFrom(pv[i],scale);
}


void xyzVector::Scale()
{
	DOUBLE_TYPE absy = fabs(at(1).value);
	DOUBLE_TYPE absz = fabs(at(2).value);
	DOUBLE_TYPE x = at(0).value;
	DOUBLE_TYPE scaleSize = max<DOUBLE_TYPE>(absy + (x * (1-2 * absy)),absz - (x * (1-2 * absz)));
	if (scaleSize >= 1)
	{
		scaleSize *= 1.00001;
		at(0).value/=scaleSize;
		at(1).value/=scaleSize;
		at(2).value/=scaleSize;
	}
};

double xyzVector::MaxPossAmplify()
{
	double MaxPoss = 1000;
	for (size_t i = 0;i < size();i++)
	{
		if (at(0).value != 0.0)
			MaxPoss = min<double>(MaxPoss,fabs(1.0/at(i).value));
	}
	return int(MaxPoss * 10.0)/10.0;
}

double xyzVectorVector::MaxPossAmplify()
{
	double MaxPoss = 1000;
	for (size_t i = 0;i < size();i++)
	{
		MaxPoss = min<double>(MaxPoss,at(i).MaxPossAmplify());
	}
	return int(MaxPoss * 10.0)/10.0;
}

xyzVector & xyzVector::operator *= (DOUBLE_TYPE a)
{
	for (size_t i = 0;i < size();i++)
		at(i).value *= a;
	return *this;
}

xyzVectorVector & xyzVectorVector::operator *= (DOUBLE_TYPE a)
{
	for (size_t i = 0;i < size();i++)
		at(i) *= a;
	return *this;
}

xyzVector & xyzVector::operator += (const xyzVector & a)
{
	for (size_t i = 0;i < size();i++)
		at(i).value += a[i].value;
	return *this;
}

xyzVectorVector & xyzVectorVector::operator += (const xyzVectorVector & a)
{
	for (size_t i = 0;i < size();i++)
		at(i) += a[i];
	return *this;
}


/**************************************************************************************
pssmScaler and pssmScalers
**************************************************************************************/

void pssmScaler::setValue(DOUBLE_TYPE v)
{
	value = v;
}

void pssmScaler::setOptimising(bool mode)
{
	optimise = mode;
}

DOUBLE_TYPE pssmScalers::value(size_t i)
{
	return at(i).value;
}

/**************************************************************************************
xyzVector the class that holds the pssm data in xyz format during optimisation
**************************************************************************************/

void pssmDataType::resize(size_t size,pssmVector pv)
{
	m_length = size;
	m_weights.resize(size,pv);
	xyzVector xyzv;
	xyzv.ConvFrom(pv);
	m_xyzWeights.resize(size,xyzv);

#ifdef WEIGHTSLOPE
	m_weightSlope.resize(size);
#endif
	m_orderedMatchesF.resize(size);
	m_orderedMatchesR.resize(size);
	for (size_t i = 0;i < m_length;i++)
	{
/*		m_weights[i].resize(4);
		if (pairedNucsMode)	//This bit may not be necessary
			m_xWeights[i].resize(3);
		else
			m_xyzWeights[i].resize(3);
*/
#ifdef WEIGHTSLOPE
		m_weightSlope[i] = 1;
#endif
	}
};

void pssmDataType::clear()
{
	m_length = 0;
	m_weights.clear();
	m_xyzWeights.clear();
};
void pssmDataType::makeComplement()
{
	for (size_t i = 0;i < m_weights.size();i++)
	{
		m_weights[i].makeComplement();
		m_xyzWeights[i].ConvFrom(m_weights[i]);
	}
};

void pssmDataType::setPositive()
{
	for (size_t i = 0;i < m_weights.size();i++)
	{
		for (size_t j = 0;j < m_weights[i].size();j++)
			if (m_weights[i][j].value < 0)
				(m_weights[i][j].value = -m_weights[i][j].value);
	}
};



void pssmDataType::push_back(const pssmVector & pv,const xyzVector & xyzv)
{
	m_weights.push_back(pv);
	m_xyzWeights.push_back(xyzv);
#ifdef WEIGHTSLOPE
	m_weightSlope.push_back(1);
#endif
	m_length++;
	m_orderedMatchesF.resize(m_length);
	m_orderedMatchesR.resize(m_length);
}

void pssmDataType::push_back(const pssmScaler & ps)
{
	m_scalers.push_back(ps);
}

void pssmDataType::operator = (const pssmDataType & pd)
{
	m_weights = pd.m_weights;
	m_xyzWeights = pd.m_xyzWeights;
#ifdef WEIGHTSLOPE
	m_weightSlope = pd.m_weightSlope;
#endif
	m_length = pd.m_length;
	m_optimising = pd.m_optimising;
}

void pssmDataBlock::resize(size_t i)
{
	vector<pssmDataType>::resize(i);
	m_gains.resize(i);
}

void pssmDataBlocks::resetWeights()
{
	m_globalWeights.resize(3);
	m_globalWeights[1].value = 1;
	m_globalWeights[2].value = 0;
	m_globalWeights.setOptimising(true);

	at(0)[0].resize(8);
}

bool pssmDataBlocks::open(const string filename,weightFileUser * bo)
{
	reset();
	std::ifstream f(filename.c_str());

	if (!f.is_open()) 
	{
		resetWeights();
		cout << "Unable to open weights file " << filename << "\n";
		return false;
	}

	mystring buffer;
	getline(f,buffer);
	const char * p;
#ifndef _DEBUG
	try
#endif
	{
		size_t block = 0;

		while (!f.eof())
		{
			//	Weights are combined with the 
			if (buffer.startsWith("[parameters",p))
			{
				bo->readWeightParameters(f,buffer);
			}
			else if (buffer.startsWith("[pssmblock",p))
			{
				read(f,buffer);
			}
			else if (buffer.startsWith("[globals]"))
			{
				size_t i = 0;
				while (getDataLine(f,buffer))
				{
					m_globalWeights.resize(i+1);
					m_globalWeights[i].value = atof(buffer);
					p = strchr(buffer,',');
					if (p)
					{
						m_globalWeights[i++].optimise = (p[1] == 'Y');
					}
				}
			}
			else
				getline(f,buffer);
		}
	}
#ifndef _DEBUG
	catch(...)
	{
		return false;
	}
#endif
	return true;
}

template <typename T,unsigned S>
	unsigned arraysize(const T (&v)[S]) { return S; }


void pssmDataBlocks::output(FILE * f,int N,...)
{

	for (size_t block = 0;block < size();block++)
	{
		fprintf(f,"[pssmblock%i]\n",block);

		for (size_t pssm = 0;pssm < at(block).size();pssm++)
		{
			if((at(block)[pssm].m_scalers.size() == 0) || (pssm == 0))
			{
				pssmVector pv;

				fprintf(f,"[pssmweights%i],%c\n",pssm,at(block)[pssm].m_optimising?'Y':'N');
				//	Only print the gain if we are not using scalers
				if (at(block)[pssm].m_scalers.size() == 0)
					fprintf(f,"gain,%f,%c\n",pow(10,at(block).m_gains[pssm].value),at(block).m_gains[pssm].optimise?'Y':'N');

				//	Convert the internal 3D format into the 4 nucleotide format
				for (size_t i = 0; i < at(block)[pssm].size();i++)
				{
					pv.ConvFrom(at(block)[pssm].m_xyzWeights[i]);

					fprintf(f,"%e,%e,%e,%e,%c\n",
						pv.absValue(0),pv.absValue(1),pv.absValue(2),pv.absValue(3),
						pv[0].optimise?'Y':'N');
				}
			}
			if (at(block)[pssm].m_scalers.size())
			{
				fprintf(f,"[pssmscalers%i],%c\n",pssm,at(block)[pssm].m_optimising?'Y':'N');
				fprintf(f,"gain,%f,%c\n",pow(10,at(block).m_gains[pssm].value),at(block).m_gains[pssm].optimise?'Y':'N');
				for (size_t i = 0; i < at(block)[pssm].m_scalers.size();i++)
					fprintf(f,"%e,%c\n",at(block)[pssm].m_scalers.value(i),
					at(block)[pssm].m_scalers[i].optimise?'Y':'N');
			}
		}
	}

	if (N)
	{
		fprintf(f,"[globals]\n");

		va_list ap;
		va_start(ap, N);
		for (size_t i = 0;i < min<size_t>(N,m_globalWeights.size()); i++)
		{
			char * p = va_arg(ap, char *);
			printItem(f,m_globalWeights[i],p);
		};
	}
}

void pssmDataBlocks::addNoiseToPssms(double scale)
{
	for (size_t block = 0;block < size();block++)
	{
		for (size_t pssm = 0;pssm < at(block).size();pssm++)
		{
			int modifyIndex = (int)((pssm * 4)/at(block).size());
			if((at(block)[pssm].m_scalers.size() == 0) || (pssm == 0))
			{
				for (size_t i = 0; i < at(block)[pssm].size();i++)
				{
#ifdef LINEAR_PSSM
					at(block)[pssm].m_weights[i][modifyIndex].value *= (1+scale);
#else
					at(block)[pssm].m_weights[i][modifyIndex].value += log(1+scale);
#endif
					at(block)[pssm].m_xyzWeights[i].ConvFrom(at(block)[pssm].m_weights[i]);
				}
			}
		}
	}
}
void pssmDataBlocks::initialise()
{
#ifdef TRIPLE_PSSM
	resize(3);
	pssmVector pvt(0.25,true);
	pssmVector pvf(0.25,false);
	xyzVector pxy;

	at(0).resize(1);
	at(0)[0].clear();
	pxy.ConvFrom(pvt);
//	for (int i = 0;i < 7;i++)
//	{
		at(0)[0].push_back(pvt,pxy);
//	}

	at(1).resize(1);
	at(1)[0].clear();
	at(1)[0].m_optimising = true;
	pxy.ConvFrom(pvf);
	for (int i = 0;i < 7;i++)
	{
		at(1)[0].push_back(pvf,pxy);
	}
	pxy.ConvFrom(pvt);
	for (int i = 0;i < 7;i++)
	{
		at(1)[0].push_back(pvt,pxy);
	}

	at(2).resize(1);
	at(2)[0].clear();
	at(2)[0].m_optimising = true;
	pxy.ConvFrom(pvt);
	for (int i = 0;i < 7;i++)
	{
		at(2)[0].push_back(pvt,pxy);
	}

	at(0).m_gains.resize(1);
	at(0).m_gains[0].optimise = false;
	at(0).m_gains[0].value = log10(0.00001);


	at(1).m_gains.resize(1);
	at(1).m_gains[0].optimise = false;
	at(1).m_gains[0].value = log10(0.3);

	at(2).m_gains.resize(1);
	at(2).m_gains[0].optimise = true;
	at(2).m_gains[0].value = log10(0.3);

#else
	resize(1);
	at(0).resize(1);
	at(0)[0].clear();
	xyzVector pxy;
	pxy.ConvFrom(pv);
	for (int i = 0;i < 8;i++)
	{
		at(0)[0].push_back(pv,pxy);
	}

	at(0).m_gains.resize(1);
	at(0).m_gains[0].optimise = false;
	at(0).m_gains[0].value = log10(0.1);

#endif
	m_globalWeights.resize(4);
	m_globalWeights[0].value = 0;
	m_globalWeights[0].optimise = false;
	m_globalWeights[1].value = 1;
	m_globalWeights[1].optimise = false;
	m_globalWeights[2].value = 0;
	m_globalWeights[2].optimise = false;
	m_globalWeights[3].value = 1;
	m_globalWeights[3].optimise = false;

}
void pssmDataBlocks::duplicate(int block,int dupPssm)
{
	/*	What we do depends on what stage we are at and how many PSSMs there are */
#ifdef TRIPLE_PSSM
	if (at(1)[0].m_optimising && !at(2)[0].m_optimising)
	{
		at(1).push_back(at(1)[0]);
		at(1).m_gains.push_back(at(1).m_gains[0]);
		at(1).m_gains[1].optimise = true;
	}
	else if (!at(2)[0].m_optimising)
	{
		//	Second stage, enable the third PSSM block
		at(2)[0].m_optimising = true;
	}
	else if ((at(0).size() == 1) && at(0)[0].m_optimising)
	{
		//	Third stage, 
		at(0).push_back(at(0)[0]);
		at(0).m_gains.push_back(at(0).m_gains[0]);
	}
	else if ((at(2).size() == 1) && at(2)[0].m_optimising)
	{
		//	Fourth stage, duplicate 3rd PSSM
		at(2).push_back(at(2)[0]);
		at(2).m_gains.push_back(at(2).m_gains[0]);
	}
	else if (dupPssm != -1)
	{
		at(block).push_back(at(block)[dupPssm]);
		at(block).m_gains.push_back(at(block).m_gains[dupPssm]);
	}
#else
	for (size_t block = 0;block < size();block++)
	{
		size_t currentNpssms = at(block).size();

		for (size_t pssm = 0;pssm < currentNpssms;pssm++)
				at(block)[pssm].setPositive();

		if (currentNpssms >1)
		{
			//Ensure that we have moved to 12 nucleutide PSSMs and that we are 
			//optimising the first and third globalWeights
			m_globalWeights[0].optimise = true;
			m_globalWeights[2].optimise = true;
			for (size_t pssm = 0;pssm < currentNpssms;pssm++)
				at(block)[pssm].resize(12,pssmVector(0.25,true));
		}
		if (dupPssm != -1)
		{
				at(block).push_back(at(block)[dupPssm]);
				at(block).m_gains.push_back(at(block).m_gains[dupPssm]);
		}
		else if (currentNpssms <= 2)
		{
			//	If there is just one PSSM then duplicate it
			for (size_t pssm = 0;pssm < currentNpssms;pssm++)
			{
				at(block).push_back(at(block)[pssm]);
				at(block).m_gains.push_back(at(block).m_gains[pssm]);
			}
		}
	}
#endif
}

void pssmDataBlocks::makeComplement()
{
	for (size_t block = 0;block < size();block++)
	{
		for (size_t pssm = 0;pssm < at(block).size();pssm++)
		{
			at(block)[pssm].makeComplement();
		}
	}
}


void pssmDataBlocks::read(std::ifstream & f,mystring & buffer)
{
	const char * p;
	size_t block = 0;
	size_t pssmW = 0;
	size_t pssmS = 0;

	while (!f.eof())
	{
		//	Weights are combined with the 
		if (buffer.startsWith("[pssmblock",p))
		{
			block = atoi(p);
			if (block >= size())
				resize(block+1);
			getline(f,buffer);
			//	Reset for the new block
			pssmW = 0;
			pssmS = 0;

		}
		else if (buffer.startsWith("[pssmweights",p))
		{
			//	If this is a new pssm vector then add its block of 
			//	data and the gain that will be associaed with it
			size_t pssm = atoi(p);
			//	Make sure the scalars are contiguous 
			//	Allows them to be editted out or copied to the end
			pssm = pssmW;

			pssmW++;
			if (pssm >= at(block).size())
				at(block).resize(pssm+1);
			if (pssm >= at(block).m_gains.size())
				at(block).m_gains.resize(pssm+1);

			at(block)[pssm].clear();

			const char * s = strchr(p,',');
			if (s)
				at(block)[pssm].m_optimising = (s[1] != 'N');

			while (getDataLine(f,buffer))
			{
				if(buffer.startsWith("gain,",p))
				{
					at(block).m_gains[pssm].value = log10(atof(p));
					p = strchr(p,',')+1;
					at(block).m_gains[pssm].optimise = (p[0] == 'Y');
				}
				else
				{
					pssmVector pw;

					p = buffer;
					for (int j = 0;(j < 4) && (p); j++)
					{
						pw.setValue(j,atof(p));
						p = strchr(p,',')+1;
					}
					if (p)
						pw.setOptimising(p[0] == 'Y');

					//	Prepare data space for use during oprimisation
					xyzVector pxyz;
					pxyz.ConvFrom(pw);
					at(block)[pssm].push_back(pw,pxyz);
				}
			}
		}
		else if (buffer.startsWith("[pssmscalers",p))
		{
			//	With scalars, there is a single PSSM weight, and 
			//	it is then multiplied by a set of different scalars, allowing
			//	a different emphasis to be placed on each position
			size_t pssm = atoi(p);

			if (m_firstScalar == -1)
			{
				if (pssm > 0)
				{
					resize(block+2);
					at(block+1).resize(1);
					at(block+1)[0] = at(block)[pssm];
					at(block).resize(pssm);
					m_firstScalar = (int)pssm;
					pssmS = m_firstScalar;
					block++;
				}
				else
					m_firstScalar = 0;
			}
			else
			{
				//	Make sure the scalars are contiguous 
				//	Allows them to be editted out
				if (pssm > pssmS)
					pssm = pssmS;
				pssmS++;
			}

			pssm -= m_firstScalar;

			if (pssm >= at(block).size())
				at(block).resize(pssm+1);
			if (pssm >= at(block).m_gains.size())
				at(block).m_gains.resize(pssm+1);

			const char * s = strchr(p,',');
			if (s)
				at(block)[pssm].m_optimising = (s[1] != 'N');


			if (pssm > 0)
			{
				at(block)[pssm] = at(block)[0];
			}

			while (getDataLine(f,buffer))
			{
				if(buffer.startsWith("gain,",p))
				{
					at(block).m_gains[pssm].value = log10(atof(p));
					p = strchr(p,',')+1;
					at(block).m_gains[pssm].optimise = (p[0] == 'Y');
				}
				else
				{
					pssmScaler ps;

					p = buffer;
					ps.setValue(atof(p));
					p = strchr(p,',')+1;
					ps.setOptimising(p[0] == 'Y');
					at(block)[pssm].push_back(ps);
					size_t i = at(block)[pssm].m_scalers.size()-1;
					at(block)[pssm].m_weights[i].ConvFrom(at(block)[0].m_xyzWeights[i],at(block)[pssm].m_scalers[i].value);
				}
			}
		}
		else
			return;
	}
}


void pssmDataBlocks::ConvertXyzToPssm()
{
	for (size_t block = 0; block < size(); block++)
	{
		for (size_t pssm = 0; pssm < at(block).size(); pssm++)
		{
			if (at(block)[pssm].m_scalers.size())
			{
				//	We take the weighted PSSM data and scale it by the scalers
				for (size_t i = 0;i < at(block)[pssm].size();i++)
				{
					at(block)[pssm].m_weights[i].ConvFrom(at(block)[0].m_xyzWeights[i],
						at(block)[pssm].m_scalers.value(i));
				}
			}
			else
			{
				//	Its a straightforward PSSM, convert and no scaling.
				for (size_t i = 0;i < at(block)[pssm].size();i++)
				{
					at(block)[pssm].m_weights[i].ConvFrom(at(block)[pssm].m_xyzWeights[i]);
				}
			}
		}
	}
}

void pssmDataBlocks::setGains()
{
	m_absGains.resize(0);
	m_maxLength = 0;

	for (size_t block = 0;block < size();block++)
	{
		for (size_t pssm = 0; pssm < at(block).size(); pssm++)
		{
			m_maxLength = max<int>((int)at(block)[pssm].size(),m_maxLength);
			m_absGains.push_back(pow((DOUBLE_TYPE)10.0,at(block).m_gains[pssm].value));
		}
	}
}

double pssmDataBlocks::calcWeight(const char * seq,chromosome_break_data & cbd,int & PSSM)
{
	static DOUBLE_TYPE log4 = log(4.0);

	double cumWeight = 0;
	int pssmCount = 0;
	double weights[100];

	for (size_t block = 0;block < size();block++)
	{

		double maxWeight = -100;
		for (size_t pssm = 0; pssm < at(block).size(); pssm++)
		{
			pssmDataType & ps(at(block)[pssm]);

			size_t pssmSize = ps.size();
			//	Use an ordered match instance assoiated with the PSSM as it will
			//	have been set to the right length, and does not need to be resized every time
			ps.m_orderedMatchesF.reset();

			weights[pssmCount] = 0;
			for (size_t j = 0;j < pssmSize; j++)
			{
				double wf;
				//	Only need to consider forward direction
				switch (seq[j])
				{
				case 'a':case 'A':
					wf = ps.m_weights[j][0].value; break;
				case 'c':case 'C':
					wf = ps.m_weights[j][1].value; break;
				case 'g':case 'G':
					wf = ps.m_weights[j][2].value; break;
				case 't':case 'T':
					wf = ps.m_weights[j][3].value; break;
				case 'N':
					wf = -100; break;
				}
				weights[pssmCount] += wf;
			}

#ifdef LINEAR_PSSM
			weights[pssmCount] -= (ps.size() * 0.25);
#else
			weights[pssmCount] += (log4 * ps.size());
			weights[pssmCount] = exp(weights[pssmCount]);
#endif

			weights[pssmCount] *= m_absGains[pssmCount];
			if (weights[pssmCount] > maxWeight)
			{
				maxWeight = weights[pssmCount];
				PSSM = pssmCount;
//				m_break_data[i][Chromosome].m_PSSM = pssmCount;
			}

			pssmCount++;
		}
		cumWeight += maxWeight;
	}

	cumWeight = m_globalWeights[0].value + (m_globalWeights[1].value * cumWeight) + 
		(m_globalWeights[2].value * cumWeight * cumWeight);

	return cumWeight;
}



size_t pssmDataBlocks::Nparameters()
{
	size_t parameters = 0;
	for (size_t block = 0; block < size();block++)
	{
		for (size_t pssm = 0; pssm < at(block).size();pssm++)
		{
			//	For the gain
			parameters += 1;

			//For the PSSM.  Each nucleotide is defined by three independant variables
			parameters += (at(block)[pssm].m_xyzWeights.size() * 3);
		}
	}
	return parameters;
}

void pssmDataBlocks::outputPssms(string & rootDir,const char * fname,double scale,bool RNA,int offset)
{
	char strCommand[MED_LINE_LENGTH];


	sprintf(strCommand, "%s \"%s*.mat\"", DEL_COM,rootDir.c_str());
	system(strCommand);
	sprintf(strCommand, "%s \"%s*.matl\"", DEL_COM,rootDir.c_str());
	system(strCommand);


	FILE * f2 = fopen((rootDir + fname + ".matl").c_str(),"w");
	if (f2 == 0) 
	{
		cout << "Failed to open " << fname <<  ".matl" << "\n";
	}

	//	Have a scale for each block
	vector<double> scales(size(),scale);

	char filename[FILENAME_MAX];

	bool scaling = false;

	if (scale == -1)
	{
		for (size_t block = 0; block < size();block++)
		{
			scale = 1000;
			for (size_t pssm = 0; pssm < at(block).size();pssm++)
			{
				scale = min<double>(scale,at(block)[pssm].m_xyzWeights.MaxPossAmplify());
			}
			scales[block] = floor((scale +1) *2.0)/2.0;
			if(scales[block] != 1)
				scaling = true;
		}
	}
	else if (scale != 1)
		scaling = true;

	if (scaling)
	{
		//	If we are scaling the PSSM then the unscaled data in m_xyzWeights is converted into the
		//	scaled data in m_weights, which is used to generate the PSSMs
		for (size_t block = 0; block < size();block++)
		{
			for (size_t pssm = 0; pssm < at(block).size();pssm++)
			{
				at(block)[pssm].m_weights.ConvFrom(at(block)[pssm].m_xyzWeights,scales[block]);
			}
		}
	}

	for (size_t block = 0; block < size();block++)
	{
		/*	In order to create a representative combined PSSM, convert the PSSMs into 
			a 3D vector and combine them*/
		DOUBLE_TYPE maxGain = 0,totGain = 0;

		for (size_t pssm = 0; pssm < at(block).size();pssm++)
		{
			DOUBLE_TYPE gain = pow(10,at(block).m_gains[pssm].value);

			totGain += gain;
			maxGain = max<DOUBLE_TYPE>(maxGain,gain);
		}

		bool outputCompositePSSM = false;
		xyzVectorVector xyzCombined(at(block)[0].size());	//For ordinary nucleotides
		pssmVectorVector pssmCombined(at(block)[0].size(),0);	//For ordinary nucleotides

		for (size_t pssm = 0; pssm < at(block).size();pssm++)
		{
			char pssmStr[100];
			if (at(block).size() == 1)
				strcpy(pssmStr,"");
			else
				sprintf(pssmStr,"%i",pssm+1);

			sprintf(filename,"%s%s%i_%i.mat",rootDir.c_str(),fname,block,pssm);
			FILE * f = fopen(filename,"w");
			if (f == 0) 
			{
				cout << "Failed to open " << filename << "\n";
			}
			else
			{
				DOUBLE_TYPE gain = pow(10,at(block).m_gains[pssm].value);
//				if (scale != 1.0)
					fprintf(f,"#Title k%s = %5.2f sf = %3.1f\n",pssmStr,gain,scales[block]);
//				else
//					fprintf(f,"#Title k%s = %5.2f\n",pssmStr,gain);

				if (RNA)
					fprintf(f,"#RNA\n");


				at(block)[pssm].m_weights.print(f,offset);


				if (at(block)[pssm].m_scalers.size() == 0)
				{
					DOUBLE_TYPE pssmScale = gain;// * scales[block];
					//	If we are not using weighted PSSMs then we create a PSSM that is an avergae
					//	of all of the PSSMs  This is done by adding and scaling the xyz vectors to create a 
					outputCompositePSSM = true;
					pssmVectorVector pvTemp;
					pvTemp.ConvFrom(at(block)[pssm].m_xyzWeights,pssmScale);
					pssmCombined += pvTemp;

					xyzVectorVector xyzTemp(at(block)[pssm].m_xyzWeights);
					xyzTemp *= pssmScale;
					xyzCombined += xyzTemp;

				}
				fclose(f);
			}
			if (f2)
				fprintf(f2,"\"%s\"\n",filename);
		}

		if (at(block).size() > 1)
		{
			//	We are doing a combined PSSM
			sprintf(filename,"%s%sAll%i.mat",rootDir.c_str(),fname,block);
			FILE * f = fopen(filename,"w");
			if (f == 0) 
			{
				cout << "Failed to open " << filename << "\n";
			}
			else
			{
				if (outputCompositePSSM)
				{
					DOUBLE_TYPE scale = xyzCombined.MaxPossAmplify();


//#define USE_PSSM
#ifdef USE_PSSM
					xyzVectorVector xyzTemp;
					xyzTemp.ConvFrom(pssmCombined);
					scale = xyzTemp.MaxPossAmplify();
//					scale = 1.0;
					pssmCombined.ConvFrom(xyzTemp,scale);
#else
					scale = xyzCombined.MaxPossAmplify();
					pssmCombined.ConvFrom(xyzCombined,scale);
#endif

//					if (scale != 1.0)
						fprintf(f,"#Title sf = %3.1f\n",scale);
					if (RNA)
						fprintf(f,"#RNA\n");


					pssmCombined.print(f,offset);

				}
				else
				{
//					if (scale != 1.0)
						fprintf(f,"#Title Gain = %5.2f sf = %3.1f\n",maxGain,scales[block]);
					if (RNA)
						fprintf(f,"#RNA\n");
					for (size_t i = 0; i < xyzCombined.size();i++)
					{
						pssmVector pssm;
						pssm.ConvFrom(at(block)[0].m_xyzWeights[i]);
						pssm.print(f);
					}
				}
				fclose(f);
			}
			if (f2)
				fprintf(f2,"\"%s\"\n",filename);
		}
	}

	if (f2)
		fclose(f2);

	FILE * fcgw = fopen((rootDir + fname + ".cgw").c_str(),"w");
	int item = 1;
	if (fcgw)
	{
		fprintf(fcgw,"[item%i]\n",item++);
		fprintf(fcgw,"type=matl\n");
		fprintf(fcgw,"item_name=%s.matl\n",fname);
		fprintf(fcgw,"file_path=%s%s.matl\n",rootDir.c_str(),fname);
		for (size_t block = 0; block < size();block++)
		{
			for (size_t pssm = 0; pssm < at(block).size();pssm++)
			{
				sprintf(filename,"%s%s%i_%i.mat",rootDir.c_str(),fname,block,pssm);

				fprintf(fcgw,"[item%i]\n",item++);
				fprintf(fcgw,"type=mat\n");
				fprintf(fcgw,"item_name=%s%i_%i.mat\n",fname,block,pssm);
				fprintf(fcgw,"file_path=%s\n",filename);
			}
		}
	}
}


void pssmDataBlocks::reset()
{
	resize(1);
	at(0).resize(1);
	at(0)[0].clear();
	m_firstScalar = -1;
}

