/*
bar.h - Header file for bar file manipulation
written by JIANG Hui, 
Institute for Computational and Mathematical Engineering, Stanford University
May, 2007 -

*/

#ifndef BAR_H
#define BAR_H

#include "type.h"
#include "cisendian.h"
#include "file_operation.h"

#define DATA_DOUBLE 0
#define DATA_FLOAT 1
#define DATA_INT 2
#define DATA_SHORT 3
#define DATA_CHAR 4
#define DATA_UINT 5
#define DATA_USHORT 6
#define DATA_UCHAR 7


#define PLUS_ONE ((float)1.000001)
#define MINUS_ONE ((float)-1.000001)


typedef union{
	double data_double;
	float data_float;
	int data_int;
	short data_short;
	char data_char;
	unsigned int data_uint;
	unsigned short data_ushort;
	unsigned char data_uchar;
} bar_column;

class bar_parameter{
public:
	uint len_name;
	string name;
	uint len_value;
	string value;
};

class bar_sequence{
public:
	uint len_seq_name;
	string name;
	uint len_group_name;
	string group_name;
	uint len_version_num;
	string version_num;
	uint num_parameter;
	vector <bar_parameter> parameters;
	uint num_data;
	vector<vector<bar_column> > data;
};

class bar{
public:
	float version;
	bool verbose;
	uint num_seq;
	vector<bar_sequence> sequences;
	uint num_col;
	vector <int> data_type;
	int size_data;
	uint num_parameter;
	vector <bar_parameter> parameters;
	uint num_data;
	F_TIME file_time; 

	long header_end;
	map<string,int> seq_map;
	vector<long> seq_pointers;

	bar();
	bool read_from_file_header(FILE *file);
	bool read_from_file_parameter(FILE *file, bar_parameter &par);
	bool read_from_file_sequence_info(FILE *file, bar_sequence &seq);
	bool read_from_file_data(FILE *file, vector<bar_column> &data);
	bool read_from_file_region(string file_name, string chr, int start, int end, vector<vector<bar_column> > &data);
};

inline bool is_bar_file(string file_name){
	if (!file_exists(file_name)) return false;

	FILE *file=fopen(file_name.c_str(), "rb");
	if (!file){
//		if (verbose) cout << "error opening bar file:" << file_name << endl;
		return false;
	}
	size_t num_read;
	char buf[65536];

//	cout << "reading bar file header.\n";
//read magic num
	if (8 != (num_read = big_endian_fread(buf, 1, 8, file))) {
//		cout << "error reading bar file: no magic\n";
		fclose(file);
		return false;
	}
	buf[8] = 0;
	if (strcmp(buf, "barr\r\n\032\n")) {
//		cout << "error reading bar file: bad magic\n";
		fclose(file);
		return false;
	}
	fclose(file);
	return true;
}
inline bar::bar(){
	verbose = false;
}
inline bool bar::read_from_file_header(FILE *file){
	size_t num_read;
	int i;
	char buf[65536];

	if (verbose) cout << "reading bar file header.\n";
//read magic num
	if (8 != (num_read = big_endian_fread(buf, 1, 8, file))) {
		cout << "error reading bar file: no magic\n";
		return false;
	}
	buf[8] = 0;
	if (strcmp(buf, "barr\r\n\032\n")) {
		cout << "error reading bar file: bad magic\n";
		return false;
	}
//read version
	if (1 != (num_read = big_endian_fread(&version, 4, 1, file))) {
		cout << "error reading bar file: no version\n";
		return false;
	}
	if (version != 1.0f && version != 2.0f) {
		cout << "error reading bar file: bad version\n";
		return false;
	}

	if (verbose) cout << "bar file version:" << (int) version << ".0" << endl;
//read number of sequences
	if (1 != (num_read = big_endian_fread(&num_seq, 4, 1, file))) {
		cout << "error reading bar file: # of sequences\n";
		return false;
	}
	if (verbose) cout << "# of sequences:" << num_seq << endl;
	sequences.resize(num_seq);
//read number of columns
	if (1 != (num_read = big_endian_fread(&num_col, 4, 1, file))) {
		cout << "error reading bar file: # of columns\n";
		return false;
	}
	if (verbose) cout << "# of columns:" << num_col << endl;
	data_type.resize(num_col);
//read data types
	if (verbose) cout << "reading data type description\n";
	size_data = 0;
	for (i = 0; i < (int)num_col; i++) {
		if (1 != (num_read = big_endian_fread(&data_type[i], 4, 1, file))) {
			cout << "error reading bar file: data type\n";
			return false;
		}
		switch (data_type[i]) {
			case DATA_DOUBLE:
				if (verbose) cout << "DOUBLE ";
				size_data += 8;
				break;
			case DATA_FLOAT:
				if (verbose) cout << "FLOAT ";
				size_data += 4;
				break;
			case DATA_INT:
				if (verbose) cout << "INT ";
				size_data += 4;
				break;
			case DATA_SHORT:
				if (verbose) cout << "SHORT ";
				size_data += 2;
				break;
			case DATA_CHAR:
				if (verbose) cout << "CHAR ";
				size_data += 1;
				break;
			case DATA_UINT:
				if (verbose) cout << "UINT ";
				size_data += 4;
				break;
			case DATA_USHORT:
				if (verbose) cout << "USHORT ";
				size_data += 2;
				break;
			case DATA_UCHAR:
				if (verbose) cout << "UCHAR ";
				size_data += 1;
				break;
			default:				
				cout << "UNSUPPORTED DATA TYPE" << endl;
				return false;;
		}
	}
	if (verbose) cout << endl;

//read number of parameters
	if (1 != (num_read = big_endian_fread(&num_parameter, 4, 1, file))) {
		cout << "error reading bar file: # of parameters " << endl;
		return false;
	}
	if (verbose) cout << "# of parameters:" << num_parameter << endl;
	parameters.resize(num_parameter);
//read parameters
	for (i = 0; i < (int)num_parameter; i++) {				
		if (!read_from_file_parameter(file, parameters[i])) {
			cout << "error reading bar file: bad parameter" << endl;
			return false;
		}
	}
	header_end = ftell(file);

	return true;
}

inline bool bar::read_from_file_parameter(FILE *file, bar_parameter &par){
	size_t num_read;
	char buf[65536];
//read parameter name
	if (1 != (num_read = big_endian_fread(&par.len_name, 4, 1, file))) {
		cout << "error reading bar file: length of name" << endl;
		return false;
	}
	if (par.len_name != (num_read = big_endian_fread(buf, 1, par.len_name, file))) {
		cout << "error reading bar file: parameter name " << endl;
		return false;
	}
	buf[par.len_name] = 0;
	par.name = buf;
//read parameter value;
	if (1 != (num_read = big_endian_fread(&par.len_value, 4, 1, file))) {
		cout << "error reading bar file: length of value" << endl;
		return false;
	}
	if (par.len_value != (num_read = big_endian_fread(buf, 1, par.len_value, file))) {
		cout << "error reading bar file: parameter value" << endl;
		return false;
	}
	buf[par.len_value] = 0;
	par.value = buf;
	if (verbose) cout << par.name << " = " << par.value << endl;
	return true;
}

inline bool bar::read_from_file_sequence_info(FILE *file, bar_sequence &seq){
	size_t num_read;
	int i;
	char buf[65536];
//read sequence name
	if (1 != (num_read = big_endian_fread(&seq.len_seq_name, 4, 1, file))) {
		cout << "error reading bar file: length of sequence name " << endl;
		return false;
	}
	if (seq.len_seq_name != (num_read = big_endian_fread(buf, 1, seq.len_seq_name, file))) {
		cout << "error reading bar file: sequence name " << endl;
		return false;
	}
	buf[seq.len_seq_name] = 0;
	seq.name = buf;
	if (verbose) cout << "name:" << buf << endl;
	if (version >= 2.0f) {
//read group name
		if (1 != (num_read = big_endian_fread(&seq.len_group_name, 4, 1, file))) {
			cout << "error reading bar file: length of group name for sequence " << endl;
			return false;
		}
		if (seq.len_group_name != (num_read = big_endian_fread(buf, 1, seq.len_group_name, file))) {
			cout << "error reading bar file: sequence group name " << endl;
			return false;
		}
		buf[seq.len_group_name] = 0;
		seq.group_name = buf;
		if (verbose) cout << "group name:" << buf << endl;
	}
//read version number
	if (1 != (num_read = big_endian_fread(&seq.len_version_num, 4, 1, file))) {
		cout << "error reading bar file: length of version number" << endl;
		return false;
	}
	if (seq.len_version_num != (num_read = big_endian_fread(buf, 1, seq.len_version_num, file))) {
		cout << "error reading bar file: sequence version number " << endl;
		return false;
	}
	buf[seq.len_version_num] = 0;
	seq.version_num = buf;
	if (verbose) cout << "version number:" << buf << endl;
	if (version >= 2.0f) {
//read number of parameters
		if (1 != (num_read = big_endian_fread(&seq.num_parameter, 4, 1, file))) {
			cout << "error reading bar file: # of parameters" << endl;
			return false;
		}
		if (verbose) cout << "# of parameters:" << seq.num_parameter << endl;
		seq.parameters.resize(seq.num_parameter);
//read parameters
		for (i = 0; i < (int)seq.num_parameter; i++) {				
			if (!read_from_file_parameter(file, seq.parameters[i])) {
				cout << "error reading bar file: bad parameter" << endl;
				return false;
			}
		}
	}
//read number of data points
	if (1 != (num_read = big_endian_fread(&seq.num_data, 4, 1, file))) {
		cout << "error reading bar file: # of data points" << endl;
		return false;
	}
	if (verbose) cout << "# of data points:" << seq.num_data << endl;
	return true;
}

inline bool bar::read_from_file_data(FILE *file, vector<bar_column> &data){
	size_t num_read;
	int k;
	data.resize(num_col);
	for (k = 0; k < (int) num_col; k++) {
		switch (data_type[k]) {
			case DATA_DOUBLE:
				if (1 != (num_read = big_endian_fread(&data[k].data_double, 8, 1, file))) {
					cout << "error reading bar file: bad data" << endl;
					return false;
				}
				break;
			case DATA_FLOAT:
				if (1 != (num_read = big_endian_fread(&data[k].data_float, 4, 1, file))) {
					cout << "error reading bar file: bad data" << endl;
					return false;
				}
				break;
			case DATA_INT:
				if (1 != (num_read = big_endian_fread(&data[k].data_int, 4, 1, file))) {
					cout << "error reading bar file: bad data" << endl;
					return false;
				}
				break;
			case DATA_SHORT:
				if (1 != (num_read = big_endian_fread(&data[k].data_short, 2, 1, file))) {
					cout << "error reading bar file: bad data" << endl;
					return false;
				}
				break;
			case DATA_CHAR:
				if (1 != (num_read = big_endian_fread(&data[k].data_char, 1, 1, file))) {
					cout << "error reading bar file: bad data" << endl;
					return false;
				}
				break;
			case DATA_UINT:
				if (1 != (num_read = big_endian_fread(&data[k].data_uint, 4, 1, file))) {
					cout << "error reading bar file: bad data" << endl;
					return false;
				}
				break;
			case DATA_USHORT:
				if (1 != (num_read = big_endian_fread(&data[k].data_ushort, 2, 1, file))) {
					cout << "error reading bar file: bad data" << endl;
					return false;
				}
				break;
			case DATA_UCHAR:
				if (1 != (num_read = big_endian_fread(&data[k].data_uchar, 1, 1, file))) {
					cout << "error reading bar file: bad data" << endl;
					return false;
				}
				break;
			default:				
				cout << "UNSUPPORTED DATA TYPE" << endl;
				return false;;
		}
	}
	return true;
}
inline bool bar::read_from_file_region(string file_name, string chr, int start, int end, 
									   vector<vector<bar_column> > &data){

	map<string,int>::iterator mi = seq_map.find(chr);
	if(mi != seq_map.end())
	{
		uint i = (*mi).second;

		FILE * file=fopen(file_name.c_str(), "rb");
		fseek(file,seq_pointers[i],SEEK_SET);
		long pos = ftell(file);
		if (pos<0) {
			cout << "error ftell" << endl;
			return false;
		}
		int low = 0, high = sequences[i].num_data - 1, value = start, mid = 0;
		vector<bar_column> temp_data;
		temp_data.resize(num_col);
		while (low <= high) {
			mid = (low + high) / 2;
			if (fseek(file, pos+size_data*mid, SEEK_SET)) {
				cout << "error fseek" << endl;
				return false;
			}
			if (!read_from_file_data(file, temp_data)) {
				cout << "error read data" << endl;
				return false;
			}
			if (temp_data[0].data_int > value)
				high = mid - 1;
			else if (temp_data[0].data_int < value)
				low = mid + 1;
			else
				break;
		}
		int j = mid;
		while (temp_data[0].data_int <= end) {
			if (temp_data[0].data_int >= start) data.push_back(temp_data);
			j++;
			if (j == (int)num_data) break;
			if (!read_from_file_data(file, temp_data)) {
				cout << "error read data" << endl;
				return false;
			}
		}
		fclose(file);
	}

	return true;
}
#endif //BAR_H
