/* ----------------------------------------------------------------------- */
/*  AffyLib.c : implementation of the affymetrix library                   */
/*  Author : Ji HongKai ; Time: 2004.08                                    */
/*  Abridged version Nigel Dyer August 2011								   */
/* ----------------------------------------------------------------------- */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "limits.h"

#include "MatrixLib.h"
#include "StringLib.h"
#include "SequenceLib.h"
#include "GenomeLib.h"
#include "AffyLib.h"

/* ----------------------------------------------------------------------- */ 
/*                            Global Variables                             */
/* ----------------------------------------------------------------------- */ 
int AFFY_IS_BIG_ENDIAN = 1;

/* ----------------------------------------------------------------------- */ 
/* Load BIG_ENDIAN and LITTLE_ENDIAN files.                                */
/* ----------------------------------------------------------------------- */ 
void reverse_buf(char *buf, int size)
{
	int i;
	char temp;

	for (i=0; i <size/2; i++)
	{
		temp = buf[i];
		buf[i] = buf[size-1-i];
		buf[size-1-i] = temp;
	}
}

size_t big_endian_fread(void *ptr, size_t size, size_t count, FILE *stream, 
						int little_endian_machine)
{
	size_t result;
	char *temp;
	int i;

	temp = (char *)calloc(count, size);
	result = fread(temp, size, count, stream);

	if(little_endian_machine == 1)
	{
		for(i=0; i<(int)count; i++)
		{
			reverse_buf(temp+(int)size*i, (int)size);
		}
	}

	memcpy(ptr, temp, size*count);
	free(temp);
	
	return result;
}

size_t big_endian_fwrite(const void *ptr, size_t size, size_t count, FILE *stream,
						 int little_endian_machine)
{
	size_t result;
	char *temp;
	int i;

	temp = (char *)calloc(count, size);
	memcpy(temp, ptr, size*count);

	if (little_endian_machine == 1)
	{
		for(i = 0; i<(int)count; i++)
		{
			reverse_buf(temp+(int)size*i, (int)size);
		}
	}
	result = fwrite(temp, size, count, stream);
	free(temp);

	return result;
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadDatabase_Affy2Refid()                                         */
/*  Load affy probe id to refid map.                                       */
/* ----------------------------------------------------------------------- */ 
struct tagStringPair **Affy_LoadDatabase_Affy2Refid(char strDatabasePath[], int *pPairNum)
{
	/* define */
	FILE *fpIn;
	struct tagStringPair *pPair,*pCurrentPair;
	struct tagStringPair *pPairList;
	struct tagStringPair **vPair = NULL;
	int nPairNum = 0;
	char strLine[LONG_LINE_LENGTH];
	char strProbe[LINE_LENGTH];
	char strRefid[LINE_LENGTH];
	int ni;
	char *chp1,*chp2;

	/* load */
	fpIn = NULL;
	fpIn = fopen(strDatabasePath, "r");
	if(fpIn == NULL)
	{
		printf("Error: Affy_LoadDatabase_Affy2Refid, cannot open mapping database file!\n");
		exit(EXIT_FAILURE);
	}

	nPairNum = 0;
	pCurrentPair = NULL;
	pPairList = NULL;
	while(fgets(strLine, LONG_LINE_LENGTH, fpIn) != NULL)
	{
		StrTrimLeft(strLine);
		StrTrimRight(strLine);
		if(strLine[0] == '\0')
			continue;

		/* loading pairs */
		chp1 = strchr(strLine, '\t');
		if(chp1 == NULL)
			continue;

		*chp1 = '\0';
		strcpy(strProbe, strLine);
		StrTrimRight(strProbe);
		if( (strcmp(strProbe, "") == 0) || (strcmp(strProbe, "---") == 0) || (strcmp(strProbe, "NA") == 0) )
			continue;
		
		chp2 = chp1+1;
		chp1 = strstr(chp2, "///");
		while(chp1 != NULL)
		{
			*chp1 = '\0';
			strcpy(strRefid, chp2);
			StrTrimLeft(strRefid);
			StrTrimRight(strRefid);

			if( (strcmp(strRefid, "") == 0) || (strcmp(strRefid, "---") == 0)
				|| (strcmp(strRefid, "NA") == 0) )
			{
			}
			else
			{
				pPair = NULL;
				pPair = StringPairCreate();
				if(pPair == NULL)
				{
					printf("Error: Affy_LoadDatabase_Affy2Refid, cannot create string pair!\n");
					exit(EXIT_FAILURE);
				}
				StringAddTail( &(pPair->m_pStr1), strProbe);
				StringAddTail( &(pPair->m_pStr2), strRefid);
				if(pPairList == NULL)
				{
					pPairList = pPair;
					pCurrentPair = pPair;
				}
				else
				{
					pCurrentPair->m_pNext = pPair;
					pCurrentPair = pPair;
				}
				nPairNum++;
			}

			chp2 = chp1+3;
			chp1 = strstr(chp2, "///");
		}

		/* process the last one */
		strcpy(strRefid, chp2);
		StrTrimLeft(strRefid);
		StrTrimRight(strRefid);

		if( (strcmp(strRefid, "") == 0) || (strcmp(strRefid, "---") == 0)
			|| (strcmp(strRefid, "NA") == 0) )
		{
		}
		else
		{
			pPair = NULL;
			pPair = StringPairCreate();
			if(pPair == NULL)
			{
				printf("Error: Affy_LoadDatabase_Affy2Refid, cannot create string pair!\n");
				exit(EXIT_FAILURE);
			}
			StringAddTail( &(pPair->m_pStr1), strProbe);
			StringAddTail( &(pPair->m_pStr2), strRefid);
			if(pPairList == NULL)
			{
				pPairList = pPair;
				pCurrentPair = pPair;
			}
			else
			{
				pCurrentPair->m_pNext = pPair;
				pCurrentPair = pPair;
			}
			nPairNum++;
		}
	}
	fclose(fpIn);

	if(nPairNum <= 0)
	{
		printf("Warning: Affy_LoadDatabase_Affy2Refid, null database!\n");
		*pPairNum = 0;
		return NULL;
	}

	vPair = NULL;
	vPair = (struct tagStringPair **)calloc(nPairNum, sizeof(struct tagStringPair *));
	if(vPair == NULL)
	{
		printf("Error: Affy_LoadDatabase_Affy2Refid, cannot organize pairs into array!\n");
		exit(EXIT_FAILURE);
	}

	ni=0;
	while(pPairList != NULL)
	{
		pPair = pPairList;
		pPairList = pPair->m_pNext;
		pPair->m_pNext = NULL;
		vPair[ni] = pPair;
		ni++;
	}

	if(ni != nPairNum)
	{
		printf("Error: Affy_LoadDatabase_Affy2Refid, pair number not match!\n");
		exit(EXIT_FAILURE);
	}

	/* return */
	*pPairNum = nPairNum;
	return vPair;
}

/*  Affy_BARSeq_Create()                                                   */
/*  create a BARSeq object.                                                */
/* ----------------------------------------------------------------------- */ 
struct tagBARSeq *Affy_BARSeq_Create()
{
	/* define */
	struct tagBARSeq *pBARSeq = NULL;

	/* create */
	pBARSeq = (struct tagBARSeq *)calloc(1, sizeof(struct tagBARSeq));
	if(pBARSeq == NULL)
	{
		printf("Error: Affy_BARSeq_Create, cannot create BARSeq object!\n");
		return NULL;
	}

	/* return */
	return pBARSeq;
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_BARSeq_Destroy()                                                  */
/*  delete a BARSeq object.                                                */
/* ----------------------------------------------------------------------- */ 
void Affy_BARSeq_Destroy(struct tagBARSeq **pBARSeq)
{
	int ni;

	if(pBARSeq != NULL)
	{
		if( (*pBARSeq) != NULL)
		{
			DeleteString((*pBARSeq)->pSeqName);
			DeleteString((*pBARSeq)->pSeqGroupName);
			DeleteString((*pBARSeq)->pSeqVersion);

			for(ni=0; ni<(*pBARSeq)->nParamNum; ni++)
			{
				DeleteString((*pBARSeq)->vParamName[ni]);
				DeleteString((*pBARSeq)->vParamValue[ni]);
			}
			free((*pBARSeq)->vParamName);
			free((*pBARSeq)->vParamValue);

			for(ni=0; ni<(*pBARSeq)->nColNum; ni++)
			{
				DestroyDoubleMatrix((*pBARSeq)->vData[ni]);
			}
			free((*pBARSeq)->vData);

			free(*pBARSeq);
			*pBARSeq = NULL;
		}
	}
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_BARData_Create()                                                  */
/*  create a BARData object.                                               */
/* ----------------------------------------------------------------------- */ 
struct tagBARData *Affy_BARData_Create()
{
	/* define */
	struct tagBARData *pBARData = NULL;

	/* create */
	pBARData = (struct tagBARData *)calloc(1, sizeof(struct tagBARData));
	if(pBARData == NULL)
	{
		printf("Error: Affy_BARData_Create, cannot create BARData object!\n");
		return NULL;
	}

	/* return */
	return pBARData;
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_BARData_Destroy()                                                 */
/*  delete a BARData object.                                               */
/* ----------------------------------------------------------------------- */ 
void Affy_BARData_Destroy(struct tagBARData **pBARData)
{
	int ni;

	if(pBARData != NULL)
	{
		if( (*pBARData) != NULL)
		{
			DestroyIntMatrix((*pBARData)->pFieldType);
			for(ni=0; ni<(*pBARData)->nParamNum; ni++)
			{
				DeleteString((*pBARData)->vParamName[ni]);
				DeleteString((*pBARData)->vParamValue[ni]);
			}
			free((*pBARData)->vParamName);
			free((*pBARData)->vParamValue);

			for(ni=0; ni<(*pBARData)->nSeqNum; ni++)
			{
				Affy_BARSeq_Destroy((*pBARData)->vSeqData+ni);
			}
			free((*pBARData)->vSeqData);

			free(*pBARData);
			*pBARData = NULL;
		}
	}
}

/* ----------------------------------------------------------------------- */ 
/*  Affy_LoadBAR_Fast()                                                    */
/*  Loading raw data from a single affymetrix's *.bar file                 */
/* ----------------------------------------------------------------------- */ 
struct tagBARData *Affy_LoadBAR_Fast(char strFileName[])
{
	/* define */
	struct tagBARData *pBARData = NULL;

	unsigned int x = 1;
	int little_endian_machine = (1 == *((char*)(&x)));
	size_t unitsize;
	char *tempmem;
	char *tempptr;
	
	FILE *fpIn;
	int ni,nj,nk;
	int nLen;

	double dV;
	float fV;
	int nV;
	short sV;
	char cV;
	unsigned int unV;
	unsigned short usV;
	unsigned char ucV;

	/* init */
	fpIn = NULL;
	fpIn = fopen(strFileName, "rb");
	if(fpIn == NULL)
	{
		printf("Error: Affy_LoadBAR, cannot open *.bar file!\n");
		exit(EXIT_FAILURE);
	}

	pBARData = Affy_BARData_Create();
	if(pBARData == NULL)
	{
		printf("Warning: Affy_LoadBAR, BARData object was not created!\n");
		exit(EXIT_FAILURE);
	}

	/* load magic number */
	if(big_endian_fread(pBARData->strMagicnumber, 1, 8, fpIn, little_endian_machine) != 8)
	{
		printf("Error: Affy_LoadBAR, cannot load magic number.\n");
		exit(EXIT_FAILURE);
	}
	pBARData->strMagicnumber[8] = '\0';
	
	if(strcmp(pBARData->strMagicnumber, "barr\r\n\032\n") != 0)
	{
		printf("Error: Affy_LoadBAR, cannot load magic number correctly.\n");
		exit(EXIT_FAILURE);
	}

	/* load version number */
	if(big_endian_fread(&(pBARData->fVersionnumber), FLOAT_SIZE, 1, fpIn, little_endian_machine) != 1)
	{
		printf("Error: Affy_LoadBAR, cannot load version number.\n");
		exit(EXIT_FAILURE);
	}

	/* load sequence number */
	if(big_endian_fread(&(pBARData->nSeqNum), INT_SIZE, 1, fpIn, little_endian_machine) != 1)
	{
		printf("Error: Affy_LoadBAR, cannot load sequence number.\n");
		exit(EXIT_FAILURE);
	}

	/* load column number */
	if(big_endian_fread(&(pBARData->nColNum), INT_SIZE, 1, fpIn, little_endian_machine) != 1)
	{
		printf("Error: Affy_LoadBAR, cannot load column number.\n");
		exit(EXIT_FAILURE);
	}

	/* load column type */
	if(pBARData->nColNum>0)
	{
		pBARData->pFieldType = CreateIntMatrix(1, pBARData->nColNum);
		if(pBARData->pFieldType == NULL)
		{
			printf("Error: Affy_LoadBAR, cannot allocate memory for field type.\n");
			exit(EXIT_FAILURE);
		}
	
		if((int)big_endian_fread(pBARData->pFieldType->pMatElement, INT_SIZE, pBARData->nColNum, fpIn, little_endian_machine) != pBARData->nColNum)
		{
			printf("Error: Affy_LoadBAR, cannot load field type correctly.\n");
			exit(EXIT_FAILURE);
		}
	}

	/* load parameter name/value pairs */
	if(big_endian_fread(&(pBARData->nParamNum), INT_SIZE, 1, fpIn, little_endian_machine) != 1)
	{
		printf("Error: Affy_LoadBAR, cannot load number of name/value pairs.\n");
		exit(EXIT_FAILURE);
	}

	if(pBARData->nParamNum > 0)
	{
		pBARData->vParamName = (struct tagString **)calloc(pBARData->nParamNum, sizeof(struct tagString *));
		pBARData->vParamValue = (struct tagString **)calloc(pBARData->nParamNum, sizeof(struct tagString *));

		if( (pBARData->vParamName == NULL) || (pBARData->vParamValue == NULL) )
		{
			printf("Error: Affy_LoadBAR, cannot allocate memory for loading parameter name/value pairs\n");
			exit(EXIT_FAILURE);
		}

		for(ni=0; ni<pBARData->nParamNum; ni++)
		{
			if(big_endian_fread(&nLen, INT_SIZE, 1, fpIn, little_endian_machine) != 1)
			{
				printf("Error: Affy_LoadBAR, cannot load parameter name length.\n");
				exit(EXIT_FAILURE);
			}

			if(nLen > 0)
			{
				pBARData->vParamName[ni] = CreateString(nLen);
				if(pBARData->vParamName[ni] == NULL)
				{
					printf("Error: Affy_LoadBAR, cannot load parameter name.\n");
					exit(EXIT_FAILURE);
				}
			
				if(big_endian_fread(pBARData->vParamName[ni]->m_pString, 1, nLen, fpIn, little_endian_machine) != nLen)
				{
					printf("Error: Affy_LoadBAR, cannot load parameter name.\n");
					exit(EXIT_FAILURE);
				}
				pBARData->vParamName[ni]->m_pString[nLen] = '\0';
			}

			if(big_endian_fread(&nLen, INT_SIZE, 1, fpIn, little_endian_machine) != 1)
			{
				printf("Error: Affy_LoadBAR, cannot load parameter value length.\n");
				exit(EXIT_FAILURE);
			}

			if(nLen > 0)
			{
				pBARData->vParamValue[ni] = CreateString(nLen);
				if(pBARData->vParamValue[ni] == NULL)
				{
					printf("Error: Affy_LoadBAR, cannot load parameter value.\n");
					exit(EXIT_FAILURE);
				}

				if(big_endian_fread(pBARData->vParamValue[ni]->m_pString, 1, nLen, fpIn, little_endian_machine) != nLen)
				{
					printf("Error: Affy_LoadBAR, cannot load parameter value.\n");
					exit(EXIT_FAILURE);
				}
				pBARData->vParamValue[ni]->m_pString[nLen] = '\0';
			}
		}
	}

	/* load sequence data */
	if(pBARData->nSeqNum <= 0)
		return pBARData;

	pBARData->vSeqData = (struct tagBARSeq **)calloc(pBARData->nSeqNum, sizeof(struct tagBARSeq *));
	if(pBARData->vSeqData == NULL)
	{
		printf("Error: Affy_LoadBAR, cannot allocate memory for loading sequence data.\n");
		exit(EXIT_FAILURE);
	}

	unitsize = 0;
	for(nj=0; nj<pBARData->nColNum; nj++)
	{
		switch(pBARData->pFieldType->pMatElement[nj])
		{
			case 0: unitsize += sizeof(double);
				break;
			case 1: unitsize += sizeof(float);
				break;
			case 2: unitsize += 4;
				break;
			case 3: unitsize += 2;
				break;
			case 4: unitsize += 1;
				break;
			case 5: unitsize += 4;
				break;
			case 6: unitsize += 2;
				break;
			case 7: unitsize += 1;
				break;
		}
	}

	for(ni=0; ni<pBARData->nSeqNum; ni++)
	{
		/* create BARSeq object */
		pBARData->vSeqData[ni] = Affy_BARSeq_Create();
		if(pBARData->vSeqData[ni] == NULL)
		{
			printf("Error: Affy_LoadBAR, cannot create BARSeq object.\n");
			exit(EXIT_FAILURE);
		}
		pBARData->vSeqData[ni]->nColNum = pBARData->nColNum;


		/* load sequence name */
		if(big_endian_fread(&nLen, INT_SIZE, 1, fpIn, little_endian_machine) != 1)
		{
			printf("Error: Affy_LoadBAR, cannot load sequence name length.\n");
			exit(EXIT_FAILURE);
		}

		if(nLen > 0)
		{
			pBARData->vSeqData[ni]->pSeqName = CreateString(nLen);
			if(pBARData->vSeqData[ni]->pSeqName == NULL)
			{
				printf("Error: Affy_LoadBAR, cannot load sequence name.\n");
				exit(EXIT_FAILURE);
			}
		
			if(big_endian_fread(pBARData->vSeqData[ni]->pSeqName->m_pString, 1, nLen, fpIn, little_endian_machine) != nLen)
			{
				printf("Error: Affy_LoadBAR, cannot load sequence name.\n");
				exit(EXIT_FAILURE);
			}
			pBARData->vSeqData[ni]->pSeqName->m_pString[nLen] = '\0';
		}

		/* load group name */
		if(pBARData->fVersionnumber > 1.5)
		{
			if(big_endian_fread(&nLen, INT_SIZE, 1, fpIn, little_endian_machine) != 1)
			{
				printf("Error: Affy_LoadBAR, cannot load sequence group name length.\n");
				exit(EXIT_FAILURE);
			}

			if(nLen > 0)
			{
				pBARData->vSeqData[ni]->pSeqGroupName = CreateString(nLen);
				if(pBARData->vSeqData[ni]->pSeqGroupName == NULL)
				{
					printf("Error: Affy_LoadBAR, cannot load sequence group name.\n");
					exit(EXIT_FAILURE);
				}
			
				if(big_endian_fread(pBARData->vSeqData[ni]->pSeqGroupName->m_pString, 1, nLen, fpIn, little_endian_machine) != nLen)
				{
					printf("Error: Affy_LoadBAR, cannot load sequence group name.\n");
					exit(EXIT_FAILURE);
				}
				pBARData->vSeqData[ni]->pSeqGroupName->m_pString[nLen] = '\0';
			}
		}

		/* load sequence version */
		if(big_endian_fread(&nLen, INT_SIZE, 1, fpIn, little_endian_machine) != 1)
		{
			printf("Error: Affy_LoadBAR, cannot load sequence version.\n");
			exit(EXIT_FAILURE);
		}

		if(nLen > 0)
		{
			pBARData->vSeqData[ni]->pSeqVersion = CreateString(nLen);
			if(pBARData->vSeqData[ni]->pSeqVersion == NULL)
			{
				printf("Error: Affy_LoadBAR, cannot load sequence version.\n");
				exit(EXIT_FAILURE);
			}
		
			if(big_endian_fread(pBARData->vSeqData[ni]->pSeqVersion->m_pString, 1, nLen, fpIn, little_endian_machine) != nLen)
			{
				printf("Error: Affy_LoadBAR, cannot load sequence version.\n");
				exit(EXIT_FAILURE);
			}
			pBARData->vSeqData[ni]->pSeqVersion->m_pString[nLen] = '\0';
		}

		/* load parameter name/value pairs */
		if(pBARData->fVersionnumber > 1.5)
		{
			if(big_endian_fread(&(pBARData->vSeqData[ni]->nParamNum), INT_SIZE, 1, fpIn, little_endian_machine) != 1)
			{
				printf("Error: Affy_LoadBAR, cannot load number of name/value pairs for BARseq object.\n");
				exit(EXIT_FAILURE);
			}

			if(pBARData->vSeqData[ni]->nParamNum > 0)
			{
				pBARData->vSeqData[ni]->vParamName = (struct tagString **)calloc(pBARData->vSeqData[ni]->nParamNum, sizeof(struct tagString *));
				pBARData->vSeqData[ni]->vParamValue = (struct tagString **)calloc(pBARData->vSeqData[ni]->nParamNum, sizeof(struct tagString *));

				if( (pBARData->vSeqData[ni]->vParamName == NULL) || (pBARData->vSeqData[ni]->vParamValue == NULL) )
				{
					printf("Error: Affy_LoadBAR, cannot allocate memory for loading sequence parameter name/value pairs\n");
					exit(EXIT_FAILURE);
				}

				for(nj=0; nj<pBARData->vSeqData[ni]->nParamNum; nj++)
				{
					if(big_endian_fread(&nLen, INT_SIZE, 1, fpIn, little_endian_machine) != 1)
					{
						printf("Error: Affy_LoadBAR, cannot load sequence parameter name length.\n");
						exit(EXIT_FAILURE);
					}

					if(nLen > 0)
					{
						pBARData->vSeqData[ni]->vParamName[nj] = CreateString(nLen);
						if(pBARData->vSeqData[ni]->vParamName[nj] == NULL)
						{
							printf("Error: Affy_LoadBAR, cannot load sequence parameter name.\n");
							exit(EXIT_FAILURE);
						}

						if(big_endian_fread(pBARData->vSeqData[ni]->vParamName[nj]->m_pString, 1, nLen, fpIn, little_endian_machine) != nLen)
						{
							printf("Error: Affy_LoadBAR, cannot load sequence parameter name.\n");
							exit(EXIT_FAILURE);
						}
						pBARData->vSeqData[ni]->vParamName[nj]->m_pString[nLen] = '\0';
					}

					if(big_endian_fread(&nLen, INT_SIZE, 1, fpIn, little_endian_machine) != 1)
					{
						printf("Error: Affy_LoadBAR, cannot load sequence parameter value length.\n");
						exit(EXIT_FAILURE);
					}

					if(nLen > 0)
					{					
						pBARData->vSeqData[ni]->vParamValue[nj] = CreateString(nLen);
						if(pBARData->vSeqData[ni]->vParamValue[nj] == NULL)
						{
							printf("Error: Affy_LoadBAR, cannot load sequence parameter value.\n");
							exit(EXIT_FAILURE);
						}

						if(big_endian_fread(pBARData->vSeqData[ni]->vParamValue[nj]->m_pString, 1, nLen, fpIn, little_endian_machine) != nLen)
						{
							printf("Error: Affy_LoadBAR, cannot load parameter value.\n");
							exit(EXIT_FAILURE);
						}
						pBARData->vSeqData[ni]->vParamValue[nj]->m_pString[nLen] = '\0';
					}
				}
			}
		}

		/* load data points */
		if(big_endian_fread(&(pBARData->vSeqData[ni]->nDataNum), INT_SIZE, 1, fpIn, little_endian_machine) != 1)
		{
			printf("Error: Affy_LoadBAR, cannot load number of data points for a sequence.\n");
			exit(EXIT_FAILURE);
		}

		if(pBARData->vSeqData[ni]->nColNum > 0)
		{
			pBARData->vSeqData[ni]->vData = (struct DOUBLEMATRIX **)calloc(pBARData->vSeqData[ni]->nColNum, sizeof(struct DOUBLEMATRIX *));
			if(pBARData->vSeqData[ni]->vData == NULL)
			{
				printf("Error: Affy_LoadBAR, cannot allocate memory for loading sequence data.\n");
				exit(EXIT_FAILURE);
			}

			if(pBARData->vSeqData[ni]->nDataNum > 0)
			{
				for(nj=0; nj<pBARData->nColNum; nj++)
				{
					pBARData->vSeqData[ni]->vData[nj] = CreateDoubleMatrix(1, pBARData->vSeqData[ni]->nDataNum);
					if(pBARData->vSeqData[ni]->vData[nj] == NULL)
					{
						printf("Error: Affy_LoadBAR, cannot allocate memory for loading sequence data.\n");
						exit(EXIT_FAILURE);
					}
				}

				tempmem = NULL;
				tempmem = (char *)calloc(pBARData->vSeqData[ni]->nDataNum, unitsize);
				if(tempmem == NULL)
				{
					printf("Error: Affy_LoadBAR, cannot allocate memory for loading sequence data.\n");
					exit(EXIT_FAILURE);
				}

				if( (int)fread(tempmem, unitsize, pBARData->vSeqData[ni]->nDataNum, fpIn) != pBARData->vSeqData[ni]->nDataNum)
				{
					printf("Error: Affy_LoadBAR, cannot load sequence data correctly.\n");
					exit(EXIT_FAILURE);
				}
				tempptr = tempmem;

				for(nk=0; nk<pBARData->vSeqData[ni]->nDataNum; nk++)
				{
					for(nj=0; nj<pBARData->nColNum; nj++)
					{
						switch(pBARData->pFieldType->pMatElement[nj])
						{
							case 0: 
								if(little_endian_machine == 1)
								{
									reverse_buf(tempptr, sizeof(double));
								}
								memcpy(&dV, tempptr, sizeof(double));
								tempptr += sizeof(double);
								pBARData->vSeqData[ni]->vData[nj]->pMatElement[nk] = dV;
								break;
							case 1: 
								if(little_endian_machine == 1)
								{
									reverse_buf(tempptr, sizeof(float));
								}
								memcpy(&fV, tempptr, sizeof(float));
								tempptr += sizeof(float);
								pBARData->vSeqData[ni]->vData[nj]->pMatElement[nk] = fV;
								break;
							case 2: 
								if(little_endian_machine == 1)
								{
									reverse_buf(tempptr, 4);
								}
								memcpy(&nV, tempptr, 4);
								tempptr += 4;
								pBARData->vSeqData[ni]->vData[nj]->pMatElement[nk] = nV;
								break;
							case 3: 
								if(little_endian_machine == 1)
								{
									reverse_buf(tempptr, 2);
								}
								memcpy(&sV, tempptr, 2);
								tempptr += 2;
								pBARData->vSeqData[ni]->vData[nj]->pMatElement[nk] = sV;
								break;
							case 4: 
								if(little_endian_machine == 1)
								{
									reverse_buf(tempptr, 1);
								}
								memcpy(&cV, tempptr, 1);
								tempptr += 1;
								pBARData->vSeqData[ni]->vData[nj]->pMatElement[nk] = cV;
								break;
							case 5: 
								if(little_endian_machine == 1)
								{
									reverse_buf(tempptr, 4);
								}
								memcpy(&unV, tempptr, 4);
								tempptr += 4;
								pBARData->vSeqData[ni]->vData[nj]->pMatElement[nk] = unV;
								break;
							case 6: 
								if(little_endian_machine == 1)
								{
									reverse_buf(tempptr, 2);
								}
								memcpy(&usV, tempptr, 2);
								tempptr += 2;
								pBARData->vSeqData[ni]->vData[nj]->pMatElement[nk] = usV;
								break;
							case 7: 
								if(little_endian_machine == 1)
								{
									reverse_buf(tempptr, 1);
								}
								memcpy(&ucV, tempptr, 1);
								tempptr += 1;
								pBARData->vSeqData[ni]->vData[nj]->pMatElement[nk] = ucV;
								break;
						}
					}
				}

				free(tempmem);
			}
		}
	}

	/* close file */
	fclose(fpIn);

	/* return */
	return pBARData;
}
/* ----------------------------------------------------------------------- */ 
/*  Affy_SaveBAR_Columns_Fast()                                            */
/*  Saving specified columns of a bar data object to a *.bar file          */
/* ----------------------------------------------------------------------- */ 
int Affy_SaveBAR_Columns_Fast(char strFileName[], struct tagBARData *pBARData,
							  struct INTMATRIX *pCol)
{
	/* define */
	unsigned int x = 1;
	int little_endian_machine = (1 == *((char*)(&x)));
	size_t unitsize;
	char *tempmem;
	char *tempptr;
	
	FILE *fpOut;
	int ni,nj,nk;
	int nLen;

	double dV;
	float fV;
	int nV;
	short sV;
	char cV;
	unsigned int unV;
	unsigned short usV;
	unsigned char ucV;

	int nNewColNum;
	int *vNewFieldType;
	int *vNewColId;
	int nColId;

	/* initial check */
	if( (pBARData == NULL) || (pCol == NULL) )
	{
		printf("Warning: empty BAR/Column data!\n");
		return PROC_SUCCESS;
	}

	/* get column number */
	nNewColNum = 0;
	for(ni=0; ni<pCol->nWidth; ni++)
	{
		if(pCol->pMatElement[ni] == 1)
		{
			nNewColNum += 1;
		}
	}
	if(nNewColNum == 0)
	{
		printf("Warning: Affy_SaveBAR_Columns_Fast, no columns need to be exported!\n");
		return PROC_SUCCESS;
	}
	
	vNewFieldType = NULL;
	vNewFieldType = (int *)calloc(nNewColNum, sizeof(int));
	vNewColId = NULL;
	vNewColId = (int *)calloc(nNewColNum, sizeof(int));
	if( (vNewFieldType == NULL) || (vNewColId == NULL) )
	{
		printf("Error: Affy_SaveBAR_Columns_Fast, cannot allocate memory for saving columns!\n");
		return PROC_SUCCESS;
	}
	nj = 0;
	unitsize = 0;
	for(ni=0; ni<pCol->nWidth; ni++)
	{
		if(pCol->pMatElement[ni] == 1)
		{
			vNewColId[nj] = ni;
			vNewFieldType[nj] = pBARData->pFieldType->pMatElement[ni];
			switch(vNewFieldType[nj])
			{
				case 0: unitsize += sizeof(double);
					break;
				case 1: unitsize += sizeof(float);
					break;
				case 2: unitsize += 4;
					break;
				case 3: unitsize += 2;
					break;
				case 4: unitsize += 1;
					break;
				case 5: unitsize += 4;
					break;
				case 6: unitsize += 2;
					break;
				case 7: unitsize += 1;
					break;
			}
			nj++;
		}
	}

	/* open file */
	fpOut = NULL;
	fpOut = fopen(strFileName, "wb");
	if(fpOut == NULL)
	{
		printf("Error: Affy_SaveBAR_Columns_Fast, cannot open output file!\n");
		return PROC_FAILURE;
	}

	/* write magic number */
	if(strcmp(pBARData->strMagicnumber, "barr\r\n\032\n") != 0)
	{
		printf("Error: Affy_SaveBAR_Columns_Fast, magic number wrong.\n");
		exit(EXIT_FAILURE);
	}

	if(big_endian_fwrite(pBARData->strMagicnumber, 1, 8, fpOut, little_endian_machine) != 8)
	{
		printf("Error: Affy_SaveBAR, cannot write magic number.\n");
		exit(EXIT_FAILURE);
	}
	
	/* write version number */
	if(big_endian_fwrite(&(pBARData->fVersionnumber), FLOAT_SIZE, 1, fpOut, little_endian_machine) != 1)
	{
		printf("Error: Affy_SaveBAR_Columns_Fast, cannot write version number.\n");
		exit(EXIT_FAILURE);
	}

	/* write sequence number */
	if(big_endian_fwrite(&(pBARData->nSeqNum), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
	{
		printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence number.\n");
		exit(EXIT_FAILURE);
	}

	/* write column number */
	if(big_endian_fwrite(&(nNewColNum), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
	{
		printf("Error: Affy_SaveBAR_Columns_Fast, cannot write column number.\n");
		exit(EXIT_FAILURE);
	}

	/* write column type */
	if(nNewColNum>0)
	{
		if((int)big_endian_fwrite(vNewFieldType, INT_SIZE, nNewColNum, fpOut, little_endian_machine) != nNewColNum)
		{
			printf("Error: Affy_SaveBAR_Columns_Fast, cannot write field type correctly.\n");
			exit(EXIT_FAILURE);
		}
	}

	/* write parameter name/value pairs */
	if(big_endian_fwrite(&(pBARData->nParamNum), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
	{
		printf("Error: Affy_SaveBAR_Columns_Fast, cannot write number of name/value pairs.\n");
		exit(EXIT_FAILURE);
	}

	if(pBARData->nParamNum > 0)
	{
		if( (pBARData->vParamName == NULL) || (pBARData->vParamValue == NULL) )
		{
			printf("Error: Affy_SaveBAR_Columns_Fast, cannot find parameter name/value pairs\n");
			exit(EXIT_FAILURE);
		}

		for(ni=0; ni<pBARData->nParamNum; ni++)
		{
			if(pBARData->vParamName[ni] == NULL)
			{
				nLen = 0;
				if(big_endian_fwrite(&nLen, INT_SIZE, 1, fpOut, little_endian_machine) != 1)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot write parameter name length.\n");
					exit(EXIT_FAILURE);
				}
			}
			else
			{
				if(big_endian_fwrite(&(pBARData->vParamName[ni]->m_nLength), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot write parameter name length.\n");
					exit(EXIT_FAILURE);
				}

				if(big_endian_fwrite(pBARData->vParamName[ni]->m_pString, 1, pBARData->vParamName[ni]->m_nLength, fpOut, little_endian_machine) != pBARData->vParamName[ni]->m_nLength)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot write parameter name.\n");
					exit(EXIT_FAILURE);
				}
			}

			if(pBARData->vParamValue[ni] == NULL)
			{
				nLen = 0;
				if(big_endian_fwrite(&nLen, INT_SIZE, 1, fpOut, little_endian_machine) != 1)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot write parameter name length.\n");
					exit(EXIT_FAILURE);
				}
			}
			else
			{
				if(big_endian_fwrite(&(pBARData->vParamValue[ni]->m_nLength), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot write parameter value length.\n");
					exit(EXIT_FAILURE);
				}

				if(big_endian_fwrite(pBARData->vParamValue[ni]->m_pString, 1, pBARData->vParamValue[ni]->m_nLength, fpOut, little_endian_machine) != pBARData->vParamValue[ni]->m_nLength)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot write parameter value.\n");
					exit(EXIT_FAILURE);
				}
			}
		}
	}

	/* write sequence data */
	for(ni=0; ni<pBARData->nSeqNum; ni++)
	{
		/* create BARSeq object */
		if(pBARData->vSeqData[ni] == NULL)
		{
			printf("Error: Affy_SaveBAR_Columns_Fast, cannot find BARSeq object.\n");
			exit(EXIT_FAILURE);
		}

		/* write sequence name */
		if(pBARData->vSeqData[ni]->pSeqName == NULL)
		{
			nLen = 0;
			if(big_endian_fwrite(&nLen, INT_SIZE, 1, fpOut, little_endian_machine) != 1)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence name length.\n");
				exit(EXIT_FAILURE);
			}
		}
		else
		{
			if(big_endian_fwrite(&(pBARData->vSeqData[ni]->pSeqName->m_nLength), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence name length.\n");
				exit(EXIT_FAILURE);
			}
			
			if(big_endian_fwrite(pBARData->vSeqData[ni]->pSeqName->m_pString, 1, pBARData->vSeqData[ni]->pSeqName->m_nLength, fpOut, little_endian_machine) != pBARData->vSeqData[ni]->pSeqName->m_nLength)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence name.\n");
				exit(EXIT_FAILURE);
			}
		}
		
		/* write group name */
		if(pBARData->fVersionnumber > 1.5)
		{
			if(pBARData->vSeqData[ni]->pSeqGroupName == NULL)
			{
				nLen = 0;
				if(big_endian_fwrite(&nLen, INT_SIZE, 1, fpOut, little_endian_machine) != 1)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence group name length.\n");
					exit(EXIT_FAILURE);
				}
			}
			else
			{
				if(big_endian_fwrite(&(pBARData->vSeqData[ni]->pSeqGroupName->m_nLength), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence group name length.\n");
					exit(EXIT_FAILURE);
				}

				if(big_endian_fwrite(pBARData->vSeqData[ni]->pSeqGroupName->m_pString, 1, pBARData->vSeqData[ni]->pSeqGroupName->m_nLength, fpOut, little_endian_machine) != pBARData->vSeqData[ni]->pSeqGroupName->m_nLength)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence group name.\n");
					exit(EXIT_FAILURE);
				}
			}
		}

		/* write sequence version */
		if(pBARData->vSeqData[ni]->pSeqVersion == NULL)
		{
			nLen = 0;
			if(big_endian_fwrite(&nLen, INT_SIZE, 1, fpOut, little_endian_machine) != 1)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence version length.\n");
				exit(EXIT_FAILURE);
			}
		}
		else
		{
			if(big_endian_fwrite(&(pBARData->vSeqData[ni]->pSeqVersion->m_nLength), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence version length.\n");
				exit(EXIT_FAILURE);
			}
			
			if(big_endian_fwrite(pBARData->vSeqData[ni]->pSeqVersion->m_pString, 1, pBARData->vSeqData[ni]->pSeqVersion->m_nLength, fpOut, little_endian_machine) != pBARData->vSeqData[ni]->pSeqVersion->m_nLength)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence version.\n");
				exit(EXIT_FAILURE);
			}
		}
		
		/* write parameter name/value pairs */
		if(pBARData->fVersionnumber > 1.5)
		{
			if(big_endian_fwrite(&(pBARData->vSeqData[ni]->nParamNum), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot write number of name/value pairs for BARseq object.\n");
				exit(EXIT_FAILURE);
			}

			if(pBARData->vSeqData[ni]->nParamNum > 0)
			{
				if( (pBARData->vSeqData[ni]->vParamName == NULL) || (pBARData->vSeqData[ni]->vParamValue == NULL) )
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot find sequence parameter name/value pairs\n");
					exit(EXIT_FAILURE);
				}

				for(nj=0; nj<pBARData->vSeqData[ni]->nParamNum; nj++)
				{
					if(pBARData->vSeqData[ni]->vParamName[nj] == NULL)
					{
						nLen = 0;
						if(big_endian_fwrite(&nLen, INT_SIZE, 1, fpOut, little_endian_machine) != 1)
						{
							printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence parameter name length.\n");
							exit(EXIT_FAILURE);
						}
					}
					else
					{
						if(big_endian_fwrite(&(pBARData->vSeqData[ni]->vParamName[nj]->m_nLength), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
						{
							printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence parameter name length.\n");
							exit(EXIT_FAILURE);
						}

						if(big_endian_fwrite(pBARData->vSeqData[ni]->vParamName[nj]->m_pString, 1, pBARData->vSeqData[ni]->vParamName[nj]->m_nLength, fpOut, little_endian_machine) != pBARData->vSeqData[ni]->vParamName[nj]->m_nLength)
						{
							printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence parameter name.\n");
							exit(EXIT_FAILURE);
						}
					}
					
					if(pBARData->vSeqData[ni]->vParamValue[nj] == NULL)
					{
						nLen = 0;
						if(big_endian_fwrite(&nLen, INT_SIZE, 1, fpOut, little_endian_machine) != 1)
						{
							printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence parameter value length.\n");
							exit(EXIT_FAILURE);
						}
					}
					else
					{
						if(big_endian_fwrite(&(pBARData->vSeqData[ni]->vParamValue[nj]->m_nLength), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
						{
							printf("Error: Affy_SaveBAR_Columns_Fast, cannot write sequence parameter value length.\n");
							exit(EXIT_FAILURE);
						}

						if(big_endian_fwrite(pBARData->vSeqData[ni]->vParamValue[nj]->m_pString, 1, pBARData->vSeqData[ni]->vParamValue[nj]->m_nLength, fpOut, little_endian_machine) != pBARData->vSeqData[ni]->vParamValue[nj]->m_nLength)
						{
							printf("Error: Affy_SaveBAR_Columns_Fast, cannot write parameter value.\n");
							exit(EXIT_FAILURE);
						}
					}
				}
			}
		}

		/* write data points */
		if(big_endian_fwrite(&(pBARData->vSeqData[ni]->nDataNum), INT_SIZE, 1, fpOut, little_endian_machine) != 1)
		{
			printf("Error: Affy_SaveBAR_Columns_Fast, cannot write number of data points for a sequence.\n");
			exit(EXIT_FAILURE);
		}

		if(pBARData->vSeqData[ni]->nColNum > 0)
		{
			if(pBARData->vSeqData[ni]->vData == NULL)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot find sequence data.\n");
				exit(EXIT_FAILURE);
			}
		}

		if(pBARData->vSeqData[ni]->nDataNum > 0)
		{
			for(nj=0; nj<nNewColNum; nj++)
			{
				nColId = vNewColId[nj];
				if(pBARData->vSeqData[ni]->vData[nColId] == NULL)
				{
					printf("Error: Affy_SaveBAR_Columns_Fast, cannot find sequence data.\n");
					exit(EXIT_FAILURE);
				}
			}

			tempmem = NULL;
			tempmem = (char *)calloc(pBARData->vSeqData[ni]->nDataNum, unitsize);
			if(tempmem == NULL)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot allocate memory for saving sequence data.\n");
				exit(EXIT_FAILURE);
			}

			tempptr = tempmem;
			for(nk=0; nk<pBARData->vSeqData[ni]->nDataNum; nk++)
			{
				for(nj=0; nj<nNewColNum; nj++)
				{
					nColId = vNewColId[nj];
					switch(vNewFieldType[nj])
					{
						case 0:
							dV = (double)(pBARData->vSeqData[ni]->vData[nColId]->pMatElement[nk]);
							memcpy(tempptr, &dV, sizeof(double));
							if(little_endian_machine == 1)
							{
								reverse_buf(tempptr, sizeof(double));
							}
							tempptr += sizeof(double);
							break;
						case 1: 
							fV = (float)(pBARData->vSeqData[ni]->vData[nColId]->pMatElement[nk]);
							memcpy(tempptr, &fV, sizeof(float));
							if(little_endian_machine == 1)
							{
								reverse_buf(tempptr, sizeof(float));
							}
							tempptr += sizeof(float);
							break;
						case 2: 
							nV = (int)(pBARData->vSeqData[ni]->vData[nColId]->pMatElement[nk]);
							memcpy(tempptr, &nV, 4);
							if(little_endian_machine == 1)
							{
								reverse_buf(tempptr, 4);
							}
							tempptr += 4;			
							break;
						case 3: 
							sV = (short)(pBARData->vSeqData[ni]->vData[nColId]->pMatElement[nk]);
							memcpy(tempptr, &sV, 2);
							if(little_endian_machine == 1)
							{
								reverse_buf(tempptr, 2);
							}
							tempptr += 2;	
							break;
						case 4: 
							cV = (char)(pBARData->vSeqData[ni]->vData[nColId]->pMatElement[nk]);
							memcpy(tempptr, &cV, 1);
							if(little_endian_machine == 1)
							{
								reverse_buf(tempptr, 1);
							}
							tempptr += 1;	
							break;
						case 5: 
							unV = (unsigned int)(pBARData->vSeqData[ni]->vData[nColId]->pMatElement[nk]);
							memcpy(tempptr, &unV, 4);
							if(little_endian_machine == 1)
							{
								reverse_buf(tempptr, 4);
							}
							tempptr += 4;
							break;
						case 6: 
							usV = (unsigned short)(pBARData->vSeqData[ni]->vData[nColId]->pMatElement[nk]);
							memcpy(tempptr, &usV, 2);
							if(little_endian_machine == 1)
							{
								reverse_buf(tempptr, 2);
							}
							tempptr += 2;	
							break;
						case 7: 
							ucV = (unsigned char)(pBARData->vSeqData[ni]->vData[nColId]->pMatElement[nk]);
							memcpy(tempptr, &ucV, 1);
							if(little_endian_machine == 1)
							{
								reverse_buf(tempptr, 1);
							}
							tempptr += 1;
							break;
					}
				}
			}

			if( (int)fwrite(tempmem, 1, unitsize*pBARData->vSeqData[ni]->nDataNum, fpOut) != unitsize*pBARData->vSeqData[ni]->nDataNum)
			{
				printf("Error: Affy_SaveBAR_Columns_Fast, cannot save sequence data correctly.\n");
				exit(EXIT_FAILURE);
			}
			free(tempmem);
		}
	}

	/* close file */
	fclose(fpOut);
	free(vNewFieldType);
	free(vNewColId);

	/* return */
	return PROC_SUCCESS;
}

