/* ----------------------------------------------------------------------- */
/*  GenomeLib.c : implementation of the genome library                     */
/*  Author : Ji HongKai ; Time: 2004.08                                    */
/*  Abridged version Nigel Dyer August 2011								   */
/* ----------------------------------------------------------------------- */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "limits.h"

#include "MatrixLib.h"
#include "StringLib.h"
#include "SequenceLib.h"
#include "GenomeLib.h"
#include "AffyLib.h"

/* ----------------------------------------------------------------------- */ 
/* ----------------------------------------------------------------------- */ 
/*  Genome_Code_4bit_GetSeq:                                               */
/*  Get sequences from Genome Sequence coding files.                       */
/*  1 byte for 2 bases.                                                    */
/*  0000: A; 0001: C; 0010: G; 0011: T.                                    */
/*  0100: a; 0101: c; 0110: g; 0111: t.                                    */
/*  1000: N;                                                               */ 
/* ----------------------------------------------------------------------- */ 
struct tagSequence *Genome_Code_4bit_GetSeq(char strInFile[], int nStart, int nEnd)
{
	/* define */
	struct tagSequence *pSeq;
	FILE *fpIn;
	int nP1,nP2,nR1,nR2;
	int numread;

	char strLine[LINE_LENGTH];
	int ni,nk,nl,nLen;
	unsigned char bBase,bChar;

	/*	Code assumes bases number from 0.   Bases number from 1 */
	nStart--;
	nEnd--;

	/* init */
	fpIn = NULL;
	fpIn = fopen(strInFile, "rb");
	if(fpIn == NULL)
	{
		printf("Error: Genome_Code_4bit_GetSeq, cannot open sequence file% s!\n",strInFile);
		return 0;
//		exit(EXIT_FAILURE);
	}
	if(nStart > nEnd)
	{
		printf("Error: Genome_Code_4bit_GetSeq, start > end!\n");
		return 0;
//		exit(EXIT_FAILURE);
	}
	nLen = nEnd-nStart+1;

	pSeq = NULL;
	pSeq = SequenceCreate();
	if(pSeq == NULL)
	{
		printf("Error: Genome_Code_4bit_GetSeq, cannot create sequences.\n");
		return 0;
//		exit(EXIT_FAILURE);
	}

	/* load seq */
	nP1 = (int)(nStart/2);
	nR1 = nStart%2;
	nP2 = (int)(nEnd/2);
	nR2 = nEnd%2;

	if( fseek( fpIn, nP1, SEEK_SET ) != 0 )
	{
		printf("Error: Genome_Code_4bit_GetSeq, cannot locate the required sequence!\n");
		exit(EXIT_FAILURE);
	}

	nk = 0;
	nl = 0;

	/* first base */
	numread = fread( &bChar, sizeof( unsigned char ), 1, fpIn );
	if(nR1 == 0)
	{
		bBase = bChar >> 4;
		switch(bBase)
		{
			case 0: strLine[nk] = 'A';
				break;
			case 1: strLine[nk] = 'C';
				break;
			case 2: strLine[nk] = 'G';
				break;
			case 3: strLine[nk] = 'T';
				break;
			case 4: strLine[nk] = 'a';
				break;
			case 5: strLine[nk] = 'c';
				break;
			case 6: strLine[nk] = 'g';
				break;
			case 7: strLine[nk] = 't';
				break;
			default: strLine[nk] = 'N';
		}
		nk++;
		nl++;
	}

	if(nl < nLen)
	{
		bBase = bChar & 0x0F;
		switch(bBase)
		{
			case 0: strLine[nk] = 'A';
				break;
			case 1: strLine[nk] = 'C';
				break;
			case 2: strLine[nk] = 'G';
				break;
			case 3: strLine[nk] = 'T';
				break;
			case 4: strLine[nk] = 'a';
				break;
			case 5: strLine[nk] = 'c';
				break;
			case 6: strLine[nk] = 'g';
				break;
			case 7: strLine[nk] = 't';
				break;
			default: strLine[nk] = 'N';
		}
		nk++;
		nl++;
	}

	/* middle bases */
	if(nl < nLen)
	{
		for(ni=(nP1+1); ni<nP2; ni++)
		{
			numread = fread( &bChar, sizeof( unsigned char ), 1, fpIn );
			bBase = bChar >> 4;
			switch(bBase)
			{
				case 0: strLine[nk] = 'A';
					break;
				case 1: strLine[nk] = 'C';
					break;
				case 2: strLine[nk] = 'G';
					break;
				case 3: strLine[nk] = 'T';
					break;
				case 4: strLine[nk] = 'a';
					break;
				case 5: strLine[nk] = 'c';
					break;
				case 6: strLine[nk] = 'g';
					break;
				case 7: strLine[nk] = 't';
					break;
				default: strLine[nk] = 'N';
			}
			nk++;
			nl++;

			if(nk == CODE_BATCH_LEN)
			{
				strLine[nk] = '\0';
				SequenceAddTail(pSeq, strLine);
				nk = 0;
			}

			bBase = bChar & 0x0F;
			switch(bBase)
			{
				case 0: strLine[nk] = 'A';
					break;
				case 1: strLine[nk] = 'C';
					break;
				case 2: strLine[nk] = 'G';
					break;
				case 3: strLine[nk] = 'T';
					break;
				case 4: strLine[nk] = 'a';
					break;
				case 5: strLine[nk] = 'c';
					break;
				case 6: strLine[nk] = 'g';
					break;
				case 7: strLine[nk] = 't';
					break;
				default: strLine[nk] = 'N';
			}
			nk++;
			nl++;

			if(nk == CODE_BATCH_LEN)
			{
				strLine[nk] = '\0';
				SequenceAddTail(pSeq, strLine);
				nk = 0;
			}
		}
	}
	
	/* last base */
	if(nl < nLen)
	{
		numread = fread( &bChar, sizeof( unsigned char ), 1, fpIn );
		bBase = bChar >> 4;
		switch(bBase)
		{
			case 0: strLine[nk] = 'A';
				break;
			case 1: strLine[nk] = 'C';
				break;
			case 2: strLine[nk] = 'G';
				break;
			case 3: strLine[nk] = 'T';
				break;
			case 4: strLine[nk] = 'a';
				break;
			case 5: strLine[nk] = 'c';
				break;
			case 6: strLine[nk] = 'g';
				break;
			case 7: strLine[nk] = 't';
				break;
			default: strLine[nk] = 'N';
		}
		nk++;
		nl++;
	
		if(nk == CODE_BATCH_LEN)
		{
			strLine[nk] = '\0';
			SequenceAddTail(pSeq, strLine);
			nk = 0;
		}
	}

	if(nl < nLen)
	{
		if(nR2 == 1)
		{
			bBase = bChar & 0x0F;
			switch(bBase)
			{
				case 0: strLine[nk] = 'A';
					break;
				case 1: strLine[nk] = 'C';
					break;
				case 2: strLine[nk] = 'G';
					break;
				case 3: strLine[nk] = 'T';
					break;
				case 4: strLine[nk] = 'a';
					break;
				case 5: strLine[nk] = 'c';
					break;
				case 6: strLine[nk] = 'g';
					break;
				case 7: strLine[nk] = 't';
					break;
				default: strLine[nk] = 'N';
			}
			nk++;
			nl++;

			if(nk == CODE_BATCH_LEN)
			{
				strLine[nk] = '\0';
				SequenceAddTail(pSeq, strLine);
				nk = 0;
			}
		}
	}

	if(nk > 0)
	{
		strLine[nk] = '\0';
		SequenceAddTail(pSeq, strLine);
		nk = 0;
	}

	/* close file */
	fclose(fpIn);

	if(nl != nLen)
	{
		printf("Error: Genome_Code_4bit_GetSeq, incorrect sequence length!\n");
		exit(EXIT_FAILURE);
	}

	/* return */
	return pSeq;
}

/* ----------------------------------------------------------------------- */ 
/*  Genome_Code_4bit_GetCompSeq:                                           */
/*  Get complement sequences from Genome Sequence coding files.            */
/*  1 byte for 2 bases.                                                    */
/*  0000: A; 0001: C; 0010: G; 0011: T.                                    */
/*  0100: a; 0101: c; 0110: g; 0111: t.                                    */
/*  1000: N;                                                               */ 
/* ----------------------------------------------------------------------- */ 
struct tagSequence *Genome_Code_4bit_GetCompSeq(char strInFile[], int nStart, int nEnd)
{
	/* define */
	struct tagSequence *pSeq;
	FILE *fpIn;
	int nP1,nP2,nR1,nR2;
	int numread;

	int ni,nk,nLen;
	unsigned char bBase,bChar;
	char * strLine;

	/*	Code assumes bases number from 0.   Bases number from 1 */
	nStart--;
	nEnd--;

	/* init */
	fpIn = NULL;
	fpIn = fopen(strInFile, "rb");
	if(fpIn == NULL)
	{
		printf("Error: Genome_Code_4bit_GetCompSeq, cannot open sequence file% s!\n",strInFile);
		return 0;
//		exit(EXIT_FAILURE);
	}
	if(nStart < nEnd)
	{
		printf("Error: Genome_Code_4bit_GetCompSeq, start < end!\n");
		return 0;
//		exit(EXIT_FAILURE);
	}
	nLen = nStart-nEnd+1;

	strLine = malloc((nLen + 1) * sizeof(char));
	strLine[nLen] = 0;

	pSeq = NULL;
	pSeq = SequenceCreate();
	if(pSeq == NULL)
	{
		printf("Error: Genome_Code_4bit_GetSeq, cannot create sequences.\n");
		return 0;
	}

	/* load seq */
	nP1 = (int)(nStart/2);
	nR1 = nStart%2;
	nP2 = (int)(nEnd/2);
	nR2 = nEnd%2;

	if( fseek( fpIn, nP2, SEEK_SET ) != 0 )
	{
		printf("Error: Genome_Code_4bit_GetSeq, cannot locate the required sequence!\n");
		exit(EXIT_FAILURE);
	}

	nk = nLen-1;

	/* first base */
	numread = fread( &bChar, sizeof( unsigned char ), 1, fpIn );
	if(nR2 == 0)
	{
		bBase = bChar >> 4;
		switch(bBase)
		{
			case 0: strLine[nk] = 'T';
				break;
			case 1: strLine[nk] = 'G';
				break;
			case 2: strLine[nk] = 'C';
				break;
			case 3: strLine[nk] = 'A';
				break;
			case 4: strLine[nk] = 't';
				break;
			case 5: strLine[nk] = 'g';
				break;
			case 6: strLine[nk] = 'c';
				break;
			case 7: strLine[nk] = 'a';
				break;
			default: strLine[nk] = 'N';
		}
		nk--;
	}

	if(nk >= 0)
	{
		bBase = bChar & 0x0F;
		switch(bBase)
		{
			case 0: strLine[nk] = 'T';
				break;
			case 1: strLine[nk] = 'G';
				break;
			case 2: strLine[nk] = 'C';
				break;
			case 3: strLine[nk] = 'A';
				break;
			case 4: strLine[nk] = 't';
				break;
			case 5: strLine[nk] = 'g';
				break;
			case 6: strLine[nk] = 'c';
				break;
			case 7: strLine[nk] = 'a';
				break;
			default: strLine[nk] = 'N';
		}
		nk--;
	}

	/* middle bases */
	if(nk >= 0)
	{
		for(ni=(nP2+1); ni<nP1; ni++)
		{
			numread = fread( &bChar, sizeof( unsigned char ), 1, fpIn );
			bBase = bChar >> 4;
			switch(bBase)
			{
			case 0: strLine[nk] = 'T';
				break;
			case 1: strLine[nk] = 'G';
				break;
			case 2: strLine[nk] = 'C';
				break;
			case 3: strLine[nk] = 'A';
				break;
			case 4: strLine[nk] = 't';
				break;
			case 5: strLine[nk] = 'g';
				break;
			case 6: strLine[nk] = 'c';
				break;
			case 7: strLine[nk] = 'a';
				break;
			default: strLine[nk] = 'N';
			}
			nk--;

			bBase = bChar & 0x0F;
			switch(bBase)
			{
			case 0: strLine[nk] = 'T';
				break;
			case 1: strLine[nk] = 'G';
				break;
			case 2: strLine[nk] = 'C';
				break;
			case 3: strLine[nk] = 'A';
				break;
			case 4: strLine[nk] = 't';
				break;
			case 5: strLine[nk] = 'g';
				break;
			case 6: strLine[nk] = 'c';
				break;
			case 7: strLine[nk] = 'a';
				break;
			default: strLine[nk] = 'N';
			}
			nk--;

		}
	}
	
	/* last base */
	if(nk >= 0)
	{
		numread = fread( &bChar, sizeof( unsigned char ), 1, fpIn );
		bBase = bChar >> 4;
		switch(bBase)
		{
			case 0: strLine[nk] = 'T';
				break;
			case 1: strLine[nk] = 'G';
				break;
			case 2: strLine[nk] = 'C';
				break;
			case 3: strLine[nk] = 'A';
				break;
			case 4: strLine[nk] = 't';
				break;
			case 5: strLine[nk] = 'g';
				break;
			case 6: strLine[nk] = 'c';
				break;
			case 7: strLine[nk] = 'a';
				break;
			default: strLine[nk] = 'N';
		}
		nk--;
	}

	if(nk >= 0)
	{
		if(nR1 == 1)
		{
			bBase = bChar & 0x0F;
			switch(bBase)
			{
			case 0: strLine[nk] = 'T';
				break;
			case 1: strLine[nk] = 'G';
				break;
			case 2: strLine[nk] = 'C';
				break;
			case 3: strLine[nk] = 'A';
				break;
			case 4: strLine[nk] = 't';
				break;
			case 5: strLine[nk] = 'g';
				break;
			case 6: strLine[nk] = 'c';
				break;
			case 7: strLine[nk] = 'a';
				break;
			default: strLine[nk] = 'N';
			}
			nk--;

		}
	}

	SequenceAddTail(pSeq, strLine);

	free(strLine);

	/* close file */
	fclose(fpIn);

	if(nk != -1)
	{
		printf("Error: Genome_Code_4bit_GetSeq, incorrect sequence length!\n");
		exit(EXIT_FAILURE);
	}

	/* return */
	return pSeq;
}

