/* ----------------------------------------------------------------------- */
/*  MatrixLib.c : implementation of the matrix process library             */
/*  Author : Ji HongKai ; Time: 1999.11                                    */
/*  Abridged version Nigel Dyer August 2011								   */
/* ----------------------------------------------------------------------- */

#include "stdlib.h"
#include "stdio.h"
#include "string.h"
#include "math.h"
#include "limits.h"
#include "time.h"

#include "StringLib.h"
#include "MatrixLib.h"
/* ----------------------------------------------------------------------- */
/* CreateIntMatrix Function                                                */
/* This function is used for creating a new INTMATRIX .                    */
/* The size of the matrix is given by two parameters : nHeight , nWidth    */
/* In other words , a nHeight*nWidth matrix will be created.               */
/* If it fails to create new matrix , the return value will be NULL .      */
/* Else the function will return the pointer to the new INTMATRIX .        */
/* ----------------------------------------------------------------------- */
struct INTMATRIX* CreateIntMatrix(long nHeight, long nWidth)
{
	/* pNewMatrix is the pointer to the new matrix. */
	struct INTMATRIX* pNewMatrix;
	/* pNewElement is the pointer to the block of matrix elements. */
	int *pNewElement;
	/* nSize is the size of matrix elements. */
	long lnSize;

	/* if nHeight=0 or nWidth=0 , return NULL */
	if((nHeight<=0)||(nWidth<=0))
		return NULL;
	/* allocates memory to struct INTMATRIX. */
	pNewMatrix = NULL;
	pNewMatrix = (struct INTMATRIX*)malloc(sizeof(struct INTMATRIX));
	if (pNewMatrix == NULL)
	{
		printf("Insufficient memory available for creating intmatrix.\n"); 
	}
	else
	{
		/* initialize new INTMATRIX. */
		pNewMatrix->nHeight = nHeight;
		pNewMatrix->nWidth = nWidth;

		/* calculate size of matrix elements. */
		lnSize = (long)(nHeight*nWidth);
		/* allocate memory for matrix elements. */
		pNewElement = NULL;
		pNewElement = (int *)calloc(lnSize, sizeof(int));
		if( pNewElement != NULL )
		{
			pNewMatrix->pMatElement = pNewElement;
		}
		else
		{
			printf( "Can't allocate memory for intmatrix elements.\n" );
			free(pNewElement);
			free(pNewMatrix);
			pNewMatrix = NULL;
		}
	}

	/* return pNewMatrix. */
	return pNewMatrix;
}

/* ----------------------------------------------------------------------- */
/* DestroyIntMatrix Function                                               */
/* This function is used for destroying a INTMATRIX .                      */
/* First, the memory block of elements will be freed .                     */
/* Then the memory block of the INTMATRIX struct will be freed .           */
/* ----------------------------------------------------------------------- */
void DestroyIntMatrix(struct INTMATRIX* pDelMatrix)
{
	/* pDelElement is the pointer to the block of matrix elements. */
	int *pDelElement;

	/* delete pDelMatrix if it exists. */
	if (pDelMatrix != NULL)
	{
		/* free memory of matrix elements. */
		pDelElement = pDelMatrix->pMatElement;
		free (pDelElement);
		/* free memory of matrix strunct. */
		free (pDelMatrix);
		pDelMatrix = NULL;
	}
} 

/* ----------------------------------------------------------------------- */
/* IMGETAT Function                                                        */
/* This function is used for getting element indicated by row index and    */
/* column index .                                                          */
/* !note: pSourceMatrix must exist .                                       */
/* Make sure you have used CreateINTMatrix function to create it           */
/* before you call this function !                                         */
/* If the row(col) indexes is larger than or equal to the row(col) number  */
/* of the matrix or smaller than 0 or some other exceptions occur, the     */
/* function will return IM_ACCESS_VIOLATION;                               */  
/* Else the function will return matrix[row][col] .                        */
/* ----------------------------------------------------------------------- */
int IMGETAT(struct INTMATRIX* pSourceMatrix,long nRow,long nCol)
{
	/* nHeight and nWidth are dimension informations of source matrix. */
	long nHeight,nWidth;
	/* pSourceElement is used for accessing the element of matrixes . */
	int *pSourceElement;
	/* nResult is the return value */
	int nResult;

	/* if source matrix doesn't exist , return IM_ACCESS_VIOLATION */
	if(pSourceMatrix==NULL)
	{
		printf("Access Violation!\n"); 
		return IM_ACCESS_VIOLATION;
	}
	/* if source matrix exists, do next */
	else
	{
		/* get dimension information */
		nHeight = pSourceMatrix->nHeight;
		nWidth = pSourceMatrix->nWidth;
		/* if nRow(nCol) < 0 or >=nHeight(nWidth),*/
		/* return DM_ACCESS_VIOLATION */
		if((nRow<0)||(nRow>=nHeight)||(nCol<0)||(nCol>=nWidth))
		{
			printf("Access Violation!\n");
			return IM_ACCESS_VIOLATION;
		}
		/* if 0<= nRow(nCol) <nHeight(nWidth), */
		/* return matrix[nRow][nCol] .  */
		else
		{
			pSourceElement = pSourceMatrix->pMatElement;
			nResult = *(pSourceElement+nRow*nWidth+nCol);
			/* return */
			return nResult;
		}
	}
}

/* ----------------------------------------------------------------------- */
/* IMLOAD Function                                                         */
/* This function is used for loading a INTMATRIX.                          */
/* If successful, return the pointer of the matrix;                        */
/* else return NULL.                                                       */
/* ----------------------------------------------------------------------- */
struct INTMATRIX* IMLOAD(char strFilePath[])
{
	/* pMatrix is used for loading matrix */
	struct INTMATRIX* pMatrix;
	/* fpLoadFile is used for loading. */
	FILE *fpLoadFile;
	/* pElement and nElement is used for access the elements of the matrix. */
	int nElement[IM_LOAD_MAX_WIDTH];
	long nElementCount;
	/* strInLine is used for loading file. */
	char strInLine[LONG_LINE_LENGTH];

	/* init */
	pMatrix = NULL;

	/* open file */
	fpLoadFile = NULL;
	fpLoadFile = fopen(strFilePath, "rt");
	if(fpLoadFile == NULL)
	{
		printf("Can't open input file!\n");
		return NULL;
	}

	/* load the first line */
	while(fgets(strInLine, LONG_LINE_LENGTH, fpLoadFile) != NULL)
	{
		/* trim right and left */
		nElementCount = IntLoadRowVector(strInLine, nElement);
		if(nElementCount == 0)
			continue;
		if(IMADDROW(&pMatrix, nElement, nElementCount) == PROC_FAILURE)
		{
			DestroyIntMatrix(pMatrix);
			pMatrix = NULL;
			printf("Can't load int matrix!\n");
			break;
		}
	}

	/* close file */
	fclose(fpLoadFile);

	/* return */
	return pMatrix;
}

/* ----------------------------------------------------------------------- */
/* IntLoadRowVector                                                        */
/* This function is used for loading a int row vector from a string.       */
/* If successful, return the count of loaded elements;                     */
/* else return 0.                                                          */
/* ----------------------------------------------------------------------- */
long IntLoadRowVector(char strInLine[], int nElement[])
{
	/* nCount is count of loaded elements. */
	long nCount;
	/* startp and endp are pointers to the string */
	char *startp;
	char *endp;
	double dTemp;
 
	/* init */
	nCount = 0;
	StrTrimRight(strInLine);
	StrTrimLeft(strInLine);
	if(strInLine[0] == '\0')
		return nCount;

	/* load */
	startp = strInLine;
	endp = strpbrk(startp, WORD_SEPARATORS);

	while(endp != NULL)
	{
		*endp = '\0';
		dTemp = atof(startp);
		nElement[nCount] = (int)dTemp;
		nCount++;
		startp = endp+1;
		StrTrimLeft(startp);
		endp = strpbrk(startp, WORD_SEPARATORS);
	}

	dTemp = atof(startp);
	nElement[nCount] = (int)dTemp;
	nCount++;

	/* return */
	return nCount;
}

/* ----------------------------------------------------------------------- */
/* IMADDROW Function                                                       */
/* This function is used for adding a row to INTMATRIX.                    */
/* If successful, return PROC_SUCCESS;                                     */
/* else return PROC_FAILURE.                                               */
/* ----------------------------------------------------------------------- */
int IMADDROW(struct INTMATRIX **pIntMatrix, int nElement[], long nElementCount)
{
	/* define */
	int *pElement;
	long lnSize;

	/* check parameter */
	if(pIntMatrix == NULL)
		return PROC_FAILURE;

	/* if pIntMatrix doesn't exist */
	if(*pIntMatrix == NULL)
	{
		*pIntMatrix = CreateIntMatrix(1, nElementCount);
		if(*pIntMatrix == NULL)
			return PROC_FAILURE;
		
		pElement = (*pIntMatrix)->pMatElement;
		memcpy(pElement, nElement, nElementCount*sizeof(int));

		return PROC_SUCCESS; 
	}

	/* if pIntMatrix does exist */
	if((*pIntMatrix)->nWidth != nElementCount)
		return PROC_FAILURE;

	lnSize = (long)(((*pIntMatrix)->nHeight)*((*pIntMatrix)->nWidth));
	pElement = (*pIntMatrix)->pMatElement;
	(*pIntMatrix)->pMatElement = (int *)realloc(pElement, (lnSize+nElementCount)*sizeof(int));
	if((*pIntMatrix)->pMatElement == NULL)
		exit(EXIT_FAILURE);

	pElement = (*pIntMatrix)->pMatElement+lnSize;
	memcpy(pElement, nElement, nElementCount*sizeof(int));
	(*pIntMatrix)->nHeight += 1;

	/* return */
	return PROC_SUCCESS;
}

/* CreateDoubleMatrix Function                                             */
/* This function is used for creating a new DOUBLEMATRIX .                 */
/* The size of the matrix is given by two parameters : nHeight , nWidth    */
/* In other words , a nHeight*nWidth matrix will be created.               */
/* If it fails to create new matrix , the return value will be NULL .      */
/* Else the function will return the pointer to the new DOUBLEMATRIX .     */
/* ----------------------------------------------------------------------- */
struct DOUBLEMATRIX* CreateDoubleMatrix(long nHeight, long nWidth)
{
	/* pNewMatrix is the pointer to the new matrix. */
	struct DOUBLEMATRIX* pNewMatrix;
	/* pNewElement is the pointer to the block of matrix elements. */
	double *pNewElement;
	/* nSize is the size of matrix elements. */
	long lnSize;

	/* if nHeight<=0 or nWidth<=0 , return NULL */
	if((nHeight<=0)||(nWidth<=0))
		return NULL;
	/* allocates memory to struct DOUBLEMATRIX. */
	pNewMatrix = NULL;
	pNewMatrix = (struct DOUBLEMATRIX*)malloc(sizeof(struct DOUBLEMATRIX));
	if (pNewMatrix == NULL)
	{
		printf("Insufficient memory available for creating double matrix.\n");
	}
	else
	{
		/* initialize new DOUBLEMATRIX. */
		pNewMatrix->nHeight = nHeight;
		pNewMatrix->nWidth = nWidth;

		/* calculate size of matrix elements. */
		lnSize = (long)(nHeight*nWidth);
		/* allocate memory for matrix elements. */
		pNewElement = NULL;
		pNewElement = (double *)calloc(lnSize, sizeof(double));
		if( pNewElement != NULL )
		{
			pNewMatrix->pMatElement = pNewElement;
		}
		else
		{
			printf( "Can't allocate memory for double matrix elements.\n" );
			free(pNewElement);
			free(pNewMatrix);
			pNewMatrix = NULL;
		}
	}

	/* return pNewMatrix. */
	return pNewMatrix;
}

/* ----------------------------------------------------------------------- */
/* DestroyDoubleMatrix Function                                            */
/* This function is used for destroying a DOUBLEMATRIX .                   */
/* First, the memory block of elements will be freed .                     */
/* Then the memory block of the DOUBLEMATRIX struct will be freed .        */
/* ----------------------------------------------------------------------- */
void DestroyDoubleMatrix(struct DOUBLEMATRIX* pDelMatrix)
{
	/* pDelElement is the pointer to the block of matrix elements. */
	double *pDelElement;

	/* delete pDelMatrix if it exists. */
	if (pDelMatrix != NULL)
	{
		/* free memory of matrix elements. */
		pDelElement = pDelMatrix->pMatElement;
		free(pDelElement);
		/* free memory of matrix strunct. */
		free (pDelMatrix);
		pDelMatrix = NULL;
	}
}

