/* ----------------------------------------------------------------------- */
/*  MatrixLib.h : interface of the matrix process library                  */
/*  Author : Ji HongKai ; Time: 1999.11                                    */
/*  Abridged version Nigel Dyer August 2011								   */
/* ----------------------------------------------------------------------- */
/* ----------------------------------------------------------------------- */
/* Declaration                                                             */
/* ----------------------------------------------------------------------- */
#define ZERO_BOUND 1e-250
#define DM_ACCESS_VIOLATION 1e250
#define IM_ACCESS_VIOLATION INT_MAX
#define LM_ACCESS_VIOLATION LONG_MAX
#define BM_ACCESS_VIOLATION 255
#define DM_LOAD_MAX_WIDTH 4096
#define LM_LOAD_MAX_WIDTH 4096
#define IM_LOAD_MAX_WIDTH 4096
#define BM_LOAD_MAX_WIDTH 4096

/* ----------------------------------------------------------------------- */
/*  Definition of INTMATRIX struct.                                        */
/*  INTMATRIX is used for manipulating matrix , the elements of which are  */
/*  of integer(int) type.                                                  */ 
/* ----------------------------------------------------------------------- */
struct INTMATRIX
{
	/* nWidth is the width of the matrix associated with pMatElement. */
	long nWidth;
	/* nHeight is the height of the matrix associated with pMatElement. */
	long nHeight;
	/* pMatElement is the pointer to the element array. */
	int * pMatElement;
};

/* ----------------------------------------------------------------------- */
/*  Definition of DOUBLEMATRIX struct.                                     */
/*  DOUBLEMATRIX is used for manipulating matrix , the elements of which   */
/*  are of double type.											           */ 
/* ----------------------------------------------------------------------- */
struct  DOUBLEMATRIX
{
	/* nWidth is the width of the matrix associated with pMatElement. */
	long nWidth;
	/* nHeight is the height of the matrix associated with pMatElement. */
	long nHeight;
	/* pMatElement is the pointer to the element array. */
	double * pMatElement;
};

/* ----------------------------------------------------------------------- */
/* CreateIntMatrix Function                                                */
/* This function is used for creating a new INTMATRIX .                    */
/* The size of the matrix is given by two parameters : nHeight , nWidth    */
/* In other words , a nHeight*nWidth matrix will be created.               */
/* If it fails to create new matrix , the return value will be NULL .      */
/* Else the function will return the pointer to the new INTMATRIX .        */
/* ----------------------------------------------------------------------- */
struct INTMATRIX* CreateIntMatrix(long nHeight, long nWidth);

/* ----------------------------------------------------------------------- */
/* DestroyIntMatrix Function                                               */
/* This function is used for destroying a INTMATRIX .                      */
/* First, the memory block of elements will be freed .                     */
/* Then the memory block of the INTMATRIX struct will be freed .           */
/* ----------------------------------------------------------------------- */
void DestroyIntMatrix(struct INTMATRIX* pDelMatrix); 

/* ----------------------------------------------------------------------- */
/* IMGETAT Function                                                        */
/* This function is used for getting element indicated by row index and    */
/* column index .                                                          */
/* !note: pSourceMatrix must exist .                                       */
/* Make sure you have used CreateINTMatrix function to create it           */
/* before you call this function !                                         */
/* If the row(col) indexes is larger than or equal to the row(col) number  */
/* of the matrix or smaller than 0 or some other exceptions occur, the     */
/* function will return INT_MAX;                                           */  
/* Else the function will return matrix[row][col] .                        */
/* ----------------------------------------------------------------------- */
int IMGETAT(struct INTMATRIX* pSourceMatrix,long nRow,long nCol); 

/* ----------------------------------------------------------------------- */
/* IMLOAD Function                                                         */
/* This function is used for loading a INTMATRIX.                          */
/* If successful, return the pointer of the matrix;                        */
/* else return NULL.                                                       */
/* ----------------------------------------------------------------------- */
struct INTMATRIX* IMLOAD(char strFilePath[]);

/* ----------------------------------------------------------------------- */
/* IntLoadRowVector                                                        */
/* This function is used for loading a int row vector from a string.       */
/* If successful, return the count of loaded elements;                     */
/* else return 0.                                                          */
/* ----------------------------------------------------------------------- */
long IntLoadRowVector(char strInLine[], int nElement[]);

/* ----------------------------------------------------------------------- */
/* IMADDROW Function                                                       */
/* This function is used for adding a row to INTMATRIX.                    */
/* If successful, return PROC_SUCCESS;                                     */
/* else return PROC_FAILURE.                                               */
/* ----------------------------------------------------------------------- */
int IMADDROW(struct INTMATRIX **pIntMatrix, int nElement[], long nElementCount);

/* ----------------------------------------------------------------------- */
/* CreateDoubleMatrix Function                                             */
/* This function is used for creating a new DOUBLEMATRIX .                 */
/* The size of the matrix is given by two parameters : nHeight , nWidth    */
/* In other words , a nHeight*nWidth matrix will be created.               */
/* If it fails to create new matrix , the return value will be NULL .      */
/* Else the function will return the pointer to the new DOUBLEMATRIX .     */
/* ----------------------------------------------------------------------- */
struct DOUBLEMATRIX* CreateDoubleMatrix(long nHeight, long nWidth);

/* ----------------------------------------------------------------------- */
/* DestroyDoubleMatrix Function                                            */
/* This function is used for destroying a DOUBLEMATRIX .                   */
/* First, the memory block of elements will be freed .                     */
/* Then the memory block of the DOUBLEMATRIX struct will be freed .        */
/* ----------------------------------------------------------------------- */
void DestroyDoubleMatrix(struct DOUBLEMATRIX* pDelMatrix);

